from collections import defaultdict
from dataclasses import dataclass
from typing import Any

import structlog

from skyvern.core.script_generations.constants import SCRIPT_TASK_BLOCKS
from skyvern.forge import app
from skyvern.schemas.workflows import BlockType
from skyvern.services import workflow_service
from skyvern.webeye.actions.action_types import ActionType
from skyvern.webeye.actions.actions import Action

LOG = structlog.get_logger(__name__)


@dataclass
class CodeGenInput:
    file_name: str
    workflow_run: dict[str, Any]
    workflow: dict[str, Any]
    workflow_blocks: list[dict[str, Any]]
    actions_by_task: dict[str, list[dict[str, Any]]]
    task_v2_child_blocks: dict[str, list[dict[str, Any]]]  # task_v2_label -> list of child blocks


def _process_action_for_block(
    action: Action,
    block_dump: dict[str, Any],
) -> dict[str, Any]:
    """Process a single action and add block-specific context like data extraction goal."""
    action_dump = action.model_dump()
    action_dump["xpath"] = action.get_xpath()
    action_dump["has_mini_agent"] = action.has_mini_agent
    if (
        "data_extraction_goal" in block_dump
        and block_dump["data_extraction_goal"]
        and action.action_type == ActionType.EXTRACT
    ):
        action_dump["data_extraction_goal"] = block_dump["data_extraction_goal"]
    if (
        "extracted_information_schema" in block_dump
        and block_dump["extracted_information_schema"]
        and action.action_type == ActionType.EXTRACT
    ):
        action_dump["data_extraction_schema"] = block_dump["extracted_information_schema"]
    return action_dump


async def transform_workflow_run_to_code_gen_input(workflow_run_id: str, organization_id: str) -> CodeGenInput:
    # get the workflow run request
    workflow_run_resp = await workflow_service.get_workflow_run_response(
        workflow_run_id=workflow_run_id, organization_id=organization_id
    )
    if not workflow_run_resp:
        raise ValueError(f"Workflow run {workflow_run_id} not found")
    run_request = workflow_run_resp.run_request
    if not run_request:
        raise ValueError(f"Workflow run {workflow_run_id} has no run request")
    workflow_run_request_json = run_request.model_dump()

    # get the workflow
    workflow = await app.WORKFLOW_SERVICE.get_workflow_by_permanent_id(
        workflow_permanent_id=run_request.workflow_id, organization_id=organization_id
    )
    if not workflow:
        raise ValueError(f"Workflow {run_request.workflow_id} not found")
    workflow_json = workflow.model_dump()

    # get the original workflow definition blocks (with templated information)
    workflow_definition_blocks = workflow.workflow_definition.blocks

    # get workflow run blocks for task execution data
    workflow_run_blocks = await app.DATABASE.get_workflow_run_blocks(
        workflow_run_id=workflow_run_id, organization_id=organization_id
    )
    workflow_run_blocks.sort(key=lambda x: x.created_at)

    # Create mapping from definition blocks by label for quick lookup
    workflow_run_blocks_by_label = {block.label: block for block in workflow_run_blocks if block.label}

    # Batch fetch all tasks and actions upfront to avoid N+1 queries
    # First pass: collect all task_ids from workflow run blocks
    all_task_ids: set[str] = set()
    for rb in workflow_run_blocks:
        if rb.block_type in SCRIPT_TASK_BLOCKS and rb.task_id:
            all_task_ids.add(rb.task_id)

    # Batch fetch all tasks and actions in 2 queries instead of N+1
    tasks_by_id: dict[str, Any] = {}
    actions_by_task_id: dict[str, list[Action]] = defaultdict(list)

    if all_task_ids:
        task_ids_list = list(all_task_ids)
        # Single query for all tasks
        tasks = await app.DATABASE.get_tasks_by_ids(task_ids=task_ids_list, organization_id=organization_id)
        tasks_by_id = {task.task_id: task for task in tasks}
        LOG.debug(
            "Batch fetched tasks for code gen",
            workflow_run_id=workflow_run_id,
            task_count=len(tasks),
        )

        # Single query for all actions (returns desc order for timeline; reverse for chronological)
        all_actions = await app.DATABASE.get_tasks_actions(task_ids=task_ids_list, organization_id=organization_id)
        all_actions.reverse()
        for action in all_actions:
            if action.task_id:
                actions_by_task_id[action.task_id].append(action)
        LOG.debug(
            "Batch fetched actions for code gen",
            workflow_run_id=workflow_run_id,
            action_count=len(all_actions),
        )

    workflow_block_dump = []
    actions_by_task: dict[str, list[dict[str, Any]]] = {}
    task_v2_child_blocks = {}

    # Loop through workflow run blocks and match to original definition blocks by label
    for definition_block in workflow_definition_blocks:
        # if definition_block.block_type == BlockType.TaskV2:
        #     raise ValueError("TaskV2 blocks are not supported yet")

        run_block = workflow_run_blocks_by_label.get(definition_block.label) if definition_block.label else None

        final_dump = {}
        if run_block:
            # Start with the original templated definition block
            final_dump = definition_block.model_dump()
        else:
            # the run_block is not a top level block - for now we will skip non top level blocks, like any blocks defined inside a loop block
            continue

        # For task blocks, add execution data while preserving templated information
        if run_block.block_type in SCRIPT_TASK_BLOCKS and run_block.task_id:
            # Use pre-fetched task data (batch fetched)
            task = tasks_by_id.get(run_block.task_id)
            if task:
                # Add task execution data but preserve original templated fields
                task_dump = task.model_dump()
                # Update with execution data, but keep templated values from definition
                if definition_block:
                    final_dump.update({k: v for k, v in task_dump.items() if k not in final_dump})
                else:
                    final_dump.update(task_dump)

                # Add run block execution metadata
                final_dump.update(
                    {
                        "task_id": run_block.task_id,
                        "status": run_block.status,
                        "output": run_block.output,
                    }
                )

                # Use pre-fetched actions (batch fetched)
                actions = actions_by_task_id.get(run_block.task_id, [])
                action_dumps = [_process_action_for_block(action, final_dump) for action in actions]
                actions_by_task[run_block.task_id] = action_dumps
            else:
                LOG.warning("Task not found", task_id=run_block.task_id)

        if run_block.block_type == BlockType.TaskV2:
            # Merge child workflow run data for task v2 blocks
            if run_block.block_workflow_run_id:
                try:
                    # Recursively get child workflow run data
                    child_code_gen_input = await transform_workflow_run_to_code_gen_input(
                        workflow_run_id=run_block.block_workflow_run_id, organization_id=organization_id
                    )

                    # Store child blocks for this task_v2 block
                    task_v2_label = run_block.label or f"task_v2_{run_block.workflow_run_block_id}"
                    task_v2_child_blocks[task_v2_label] = child_code_gen_input.workflow_blocks

                    # Merge child workflow blocks into the current workflow_block_dump (but mark them as child blocks)
                    # for child_block in child_code_gen_input.workflow_blocks:
                    #     child_block["parent_task_v2_label"] = task_v2_label
                    #     workflow_block_dump.append(child_block)

                    # Merge child actions_by_task into current actions_by_task
                    for task_id, child_actions in child_code_gen_input.actions_by_task.items():
                        actions_by_task[task_id] = child_actions

                    # Also merge nested task_v2 child blocks if any
                    for nested_label, nested_blocks in child_code_gen_input.task_v2_child_blocks.items():
                        task_v2_child_blocks[nested_label] = nested_blocks

                except Exception as e:
                    LOG.warning(
                        "Failed to merge child workflow run data for task v2 block",
                        task_v2_workflow_run_id=run_block.block_workflow_run_id,
                        error=str(e),
                    )
            else:
                LOG.warning(
                    "Task v2 block does not have a child workflow run id",
                    task_v2_label=run_block.label,
                )

        if run_block.block_type == BlockType.FOR_LOOP:
            # Process ForLoop child blocks to get actions for task blocks inside the loop
            # Child blocks have parent_workflow_run_block_id pointing to the ForLoop's workflow_run_block_id
            child_run_blocks = [
                b for b in workflow_run_blocks if b.parent_workflow_run_block_id == run_block.workflow_run_block_id
            ]
            # Create mapping of child run blocks by label
            child_run_blocks_by_label = {b.label: b for b in child_run_blocks if b.label}

            # Warn about any unlabeled child blocks that won't be matched
            unlabeled_children = [b for b in child_run_blocks if not b.label]
            if unlabeled_children:
                LOG.warning(
                    "ForLoop has child blocks without labels - these will not be matched to loop_blocks definitions",
                    forloop_label=run_block.label,
                    unlabeled_count=len(unlabeled_children),
                )

            # Get loop_blocks from the definition block
            loop_blocks = final_dump.get("loop_blocks", [])

            if loop_blocks and not child_run_blocks:
                LOG.warning(
                    "ForLoop block has loop_blocks definitions but no child run blocks found",
                    forloop_label=run_block.label,
                    workflow_run_block_id=run_block.workflow_run_block_id,
                    loop_blocks_count=len(loop_blocks),
                )
            updated_loop_blocks = []

            for loop_block_def in loop_blocks:
                loop_block_dump = loop_block_def.copy() if isinstance(loop_block_def, dict) else loop_block_def
                loop_block_label = loop_block_dump.get("label")

                # Find matching child run block
                child_run_block = child_run_blocks_by_label.get(loop_block_label) if loop_block_label else None

                if child_run_block and child_run_block.block_type in SCRIPT_TASK_BLOCKS and child_run_block.task_id:
                    # Use pre-fetched task data (batch fetched)
                    task = tasks_by_id.get(child_run_block.task_id)
                    if task:
                        task_dump = task.model_dump()
                        loop_block_dump.update({k: v for k, v in task_dump.items() if k not in loop_block_dump})
                        loop_block_dump.update(
                            {
                                "task_id": child_run_block.task_id,
                                "status": child_run_block.status,
                                "output": child_run_block.output,
                            }
                        )

                        # Use pre-fetched actions (batch fetched)
                        actions = actions_by_task_id.get(child_run_block.task_id, [])
                        action_dumps = [_process_action_for_block(action, loop_block_dump) for action in actions]
                        actions_by_task[child_run_block.task_id] = action_dumps
                    else:
                        LOG.warning(
                            "Task not found for ForLoop child block",
                            task_id=child_run_block.task_id,
                            forloop_label=run_block.label,
                        )

                updated_loop_blocks.append(loop_block_dump)

            # Update final_dump with the processed loop_blocks
            final_dump["loop_blocks"] = updated_loop_blocks

        final_dump["workflow_run_id"] = workflow_run_id
        if run_block:
            final_dump["workflow_run_block_id"] = run_block.workflow_run_block_id
        else:
            final_dump["workflow_run_block_id"] = None
        workflow_block_dump.append(final_dump)

    return CodeGenInput(
        file_name=f"{workflow_run_id}.py",
        workflow_run=workflow_run_request_json,
        workflow=workflow_json,
        workflow_blocks=workflow_block_dump,
        actions_by_task=actions_by_task,
        task_v2_child_blocks=task_v2_child_blocks,
    )



"""add browser session history index

Revision ID: 24bd4ab81e07
Revises: 1001a00ed620
Create Date: 2025-12-02 17:17:28.644094+00:00

"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "24bd4ab81e07"
down_revision: Union[str, None] = "1001a00ed620"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(
        "idx_persistent_browser_sessions_org_created_started_completed",
        "persistent_browser_sessions",
        ["organization_id", "created_at", "started_at", "completed_at"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "idx_persistent_browser_sessions_org_created_started_completed", table_name="persistent_browser_sessions"
    )
    # ### end Alembic commands ###



# This file was auto-generated by Fern from our API Definition.

import typing

ContextParameterValue = typing.Union[
    str, int, float, bool, typing.Dict[str, typing.Optional[typing.Any]], typing.List[typing.Optional[typing.Any]]
]



from typing import TYPE_CHECKING, Any

import structlog
from playwright.async_api import Page

from skyvern.client import (
    RunSdkActionRequestAction_AiAct,
    RunSdkActionRequestAction_AiClick,
    RunSdkActionRequestAction_AiInputText,
    RunSdkActionRequestAction_AiSelectOption,
    RunSdkActionRequestAction_AiUploadFile,
    RunSdkActionRequestAction_Extract,
    RunSdkActionRequestAction_LocateElement,
    RunSdkActionRequestAction_Prompt,
    RunSdkActionRequestAction_Validate,
)
from skyvern.config import settings
from skyvern.core.script_generations.skyvern_page_ai import SkyvernPageAi

if TYPE_CHECKING:
    from skyvern.library.skyvern_browser import SkyvernBrowser

LOG = structlog.get_logger()


class SdkSkyvernPageAi(SkyvernPageAi):
    """Implementation of SkyvernPageAi that makes API calls to the server."""

    def __init__(
        self,
        browser: "SkyvernBrowser",
        page: Page,
    ):
        self._browser = browser
        self._page = page

    async def ai_click(
        self,
        selector: str | None,
        intention: str,
        data: str | dict[str, Any] | None = None,
        timeout: float = settings.BROWSER_ACTION_TIMEOUT_MS,
    ) -> str | None:
        """Click an element using AI via API call."""

        LOG.info("AI click", intention=intention, workflow_run_id=self._browser.workflow_run_id)

        response = await self._browser.skyvern.run_sdk_action(
            url=self._page.url,
            browser_session_id=self._browser.browser_session_id,
            browser_address=self._browser.browser_address,
            workflow_run_id=self._browser.workflow_run_id,
            action=RunSdkActionRequestAction_AiClick(
                selector=selector,
                intention=intention,
                data=data,
                timeout=timeout,
            ),
        )
        self._browser.workflow_run_id = response.workflow_run_id
        return response.result if response.result else selector

    async def ai_input_text(
        self,
        selector: str | None,
        value: str | None,
        intention: str,
        data: str | dict[str, Any] | None = None,
        totp_identifier: str | None = None,
        totp_url: str | None = None,
        timeout: float = settings.BROWSER_ACTION_TIMEOUT_MS,
    ) -> str:
        """Input text into an element using AI via API call."""

        LOG.info("AI input text", intention=intention, workflow_run_id=self._browser.workflow_run_id)

        response = await self._browser.skyvern.run_sdk_action(
            url=self._page.url,
            action=RunSdkActionRequestAction_AiInputText(
                selector=selector,
                value=value,
                intention=intention,
                data=data,
                totp_identifier=totp_identifier,
                totp_url=totp_url,
                timeout=timeout,
            ),
            browser_session_id=self._browser.browser_session_id,
            browser_address=self._browser.browser_address,
            workflow_run_id=self._browser.workflow_run_id,
        )
        self._browser.workflow_run_id = response.workflow_run_id
        return response.result if response.result else value or ""

    async def ai_select_option(
        self,
        selector: str | None,
        value: str | None,
        intention: str,
        data: str | dict[str, Any] | None = None,
        timeout: float = settings.BROWSER_ACTION_TIMEOUT_MS,
    ) -> str:
        """Select an option from a dropdown using AI via API call."""

        LOG.info("AI select option", intention=intention, workflow_run_id=self._browser.workflow_run_id)

        response = await self._browser.skyvern.run_sdk_action(
            url=self._page.url,
            action=RunSdkActionRequestAction_AiSelectOption(
                selector=selector,
                value=value,
                intention=intention,
                data=data,
                timeout=timeout,
            ),
            browser_session_id=self._browser.browser_session_id,
            browser_address=self._browser.browser_address,
            workflow_run_id=self._browser.workflow_run_id,
        )
        self._browser.workflow_run_id = response.workflow_run_id
        return response.result if response.result else value or ""

    async def ai_upload_file(
        self,
        selector: str | None,
        files: str | None,
        intention: str,
        data: str | dict[str, Any] | None = None,
        timeout: float = settings.BROWSER_ACTION_TIMEOUT_MS,
        public_url_only: bool = False,
    ) -> str:
        """Upload a file using AI via API call."""

        LOG.info("AI upload file", intention=intention, workflow_run_id=self._browser.workflow_run_id)

        response = await self._browser.skyvern.run_sdk_action(
            url=self._page.url,
            action=RunSdkActionRequestAction_AiUploadFile(
                selector=selector,
                file_url=files,
                intention=intention,
                data=data,
                timeout=timeout,
            ),
            browser_session_id=self._browser.browser_session_id,
            browser_address=self._browser.browser_address,
            workflow_run_id=self._browser.workflow_run_id,
        )
        self._browser.workflow_run_id = response.workflow_run_id
        return response.result if response.result else files or ""

    async def ai_extract(
        self,
        prompt: str,
        schema: dict[str, Any] | list | str | None = None,
        error_code_mapping: dict[str, str] | None = None,
        intention: str | None = None,
        data: str | dict[str, Any] | None = None,
    ) -> dict[str, Any] | list | str | None:
        """Extract information from the page using AI via API call."""

        LOG.info("AI extract", prompt=prompt, workflow_run_id=self._browser.workflow_run_id)

        response = await self._browser.skyvern.run_sdk_action(
            url=self._page.url,
            action=RunSdkActionRequestAction_Extract(
                prompt=prompt,
                extract_schema=schema,
                error_code_mapping=error_code_mapping,
                intention=intention,
                data=data,
            ),
            browser_session_id=self._browser.browser_session_id,
            browser_address=self._browser.browser_address,
            workflow_run_id=self._browser.workflow_run_id,
        )
        self._browser.workflow_run_id = response.workflow_run_id
        return response.result if response.result else None

    async def ai_validate(
        self,
        prompt: str,
        model: dict[str, Any] | None = None,
    ) -> bool:
        """Validate the current page state using AI via API call."""

        LOG.info(
            "AI validate",
            prompt=prompt,
            model=model,
            workflow_run_id=self._browser.workflow_run_id,
        )

        response = await self._browser.skyvern.run_sdk_action(
            url=self._page.url,
            action=RunSdkActionRequestAction_Validate(
                prompt=prompt,
                model=model,
            ),
            browser_session_id=self._browser.browser_session_id,
            browser_address=self._browser.browser_address,
            workflow_run_id=self._browser.workflow_run_id,
        )
        self._browser.workflow_run_id = response.workflow_run_id

        return bool(response.result) if response.result is not None else False

    async def ai_act(
        self,
        prompt: str,
    ) -> None:
        """Perform an action on the page using AI via API call."""

        LOG.info("AI act", prompt=prompt, workflow_run_id=self._browser.workflow_run_id)

        response = await self._browser.skyvern.run_sdk_action(
            url=self._page.url,
            action=RunSdkActionRequestAction_AiAct(
                intention=prompt,
            ),
            browser_session_id=self._browser.browser_session_id,
            browser_address=self._browser.browser_address,
            workflow_run_id=self._browser.workflow_run_id,
        )
        self._browser.workflow_run_id = response.workflow_run_id

    async def ai_locate_element(
        self,
        prompt: str,
    ) -> str | None:
        """Locate an element on the page using AI and return its XPath selector via API call.

        Args:
            prompt: Natural language description of the element to locate (e.g., 'find "download invoices" button')

        Returns:
            XPath selector string (e.g., 'xpath=//button[@id="download"]') or None if not found
        """

        LOG.info("AI locate element", prompt=prompt, workflow_run_id=self._browser.workflow_run_id)

        response = await self._browser.skyvern.run_sdk_action(
            url=self._page.url,
            action=RunSdkActionRequestAction_LocateElement(
                prompt=prompt,
            ),
            browser_session_id=self._browser.browser_session_id,
            browser_address=self._browser.browser_address,
            workflow_run_id=self._browser.workflow_run_id,
        )
        self._browser.workflow_run_id = response.workflow_run_id

        # Return the XPath result directly
        if response.result and isinstance(response.result, str):
            return response.result

        return None

    async def ai_prompt(
        self,
        prompt: str,
        schema: dict[str, Any] | None = None,
        model: dict[str, Any] | None = None,
    ) -> dict[str, Any] | list | str | None:
        """Send a prompt to the LLM and get a response based on the provided schema via API call."""

        LOG.info(
            "AI prompt",
            prompt=prompt,
            model=model,
            workflow_run_id=self._browser.workflow_run_id,
        )

        response = await self._browser.skyvern.run_sdk_action(
            url=self._page.url,
            action=RunSdkActionRequestAction_Prompt(
                prompt=prompt,
                response_schema=schema,
                model=model,
            ),
            browser_session_id=self._browser.browser_session_id,
            browser_address=self._browser.browser_address,
            workflow_run_id=self._browser.workflow_run_id,
        )
        self._browser.workflow_run_id = response.workflow_run_id

        return response.result if response.result is not None else None



from __future__ import annotations

import hashlib
import os


def _resolve_api_key_hash_iterations() -> int:
    raw = os.environ.get("SKYVERN_MCP_API_KEY_HASH_ITERATIONS", "120000")
    try:
        return max(10_000, int(raw))
    except ValueError:
        return 120_000


_API_KEY_HASH_ITERATIONS = _resolve_api_key_hash_iterations()
_API_KEY_HASH_SALT = os.environ.get(
    "SKYVERN_MCP_API_KEY_HASH_SALT",
    "skyvern-mcp-api-key-cache-v1",
).encode("utf-8")


def hash_api_key_for_cache(api_key: str) -> str:
    """Derive a deterministic, non-reversible fingerprint for API-key keyed caches."""
    return hashlib.pbkdf2_hmac(
        "sha256",
        api_key.encode("utf-8"),
        _API_KEY_HASH_SALT,
        _API_KEY_HASH_ITERATIONS,
    ).hex()



"""db migration script

Revision ID: 8998d998feed
Revises: f78486c3f895
Create Date: 2025-09-15 05:40:16.200764+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "8998d998feed"
down_revision: Union[str, None] = "f78486c3f895"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflow_runs", sa.Column("script_run", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflow_runs", "script_run")
    # ### end Alembic commands ###



import typing
from typing import Any

if typing.TYPE_CHECKING:
    from skyvern.library.skyvern import Skyvern  # noqa: E402

# noinspection PyUnresolvedReferences
__all__ = ["Skyvern"]


def __getattr__(name: str) -> Any:
    """Lazily import Skyvern."""
    if name == "Skyvern":
        from skyvern.library.skyvern import Skyvern  # noqa: PLC0415

        globals()["Skyvern"] = Skyvern
        return Skyvern
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""
A types module for browser recording actions and events.
"""

import enum
import typing as t
from typing import Literal

from pydantic import BaseModel

from skyvern.client.types.workflow_definition_yaml_blocks_item import (
    WorkflowDefinitionYamlBlocksItem_Action,
    WorkflowDefinitionYamlBlocksItem_GotoUrl,
    WorkflowDefinitionYamlBlocksItem_Wait,
)


class ActionKind(enum.StrEnum):
    CLICK = "click"
    HOVER = "hover"
    INPUT_TEXT = "input_text"
    URL_CHANGE = "url_change"
    WAIT = "wait"


class ActionBase(BaseModel):
    kind: ActionKind
    # --
    target: "ActionTarget"
    timestamp_start: float
    timestamp_end: float
    url: str


class ActionClick(ActionBase):
    kind: t.Literal[ActionKind.CLICK]


class ActionHover(ActionBase):
    kind: t.Literal[ActionKind.HOVER]
    # --
    DURATION_THRESHOLD_MS: t.ClassVar[int] = 2000
    MIN_DURATION_THRESHOLD_MS: t.ClassVar[int] = 1000


class ActionInputText(ActionBase):
    kind: t.Literal[ActionKind.INPUT_TEXT]
    # --
    input_value: str


class ActionUrlChange(ActionBase):
    kind: t.Literal[ActionKind.URL_CHANGE]


class ActionWait(ActionBase):
    kind: t.Literal[ActionKind.WAIT]
    # --
    duration_ms: int
    MIN_DURATION_THRESHOLD_MS: t.ClassVar[int] = 5000


Action = ActionClick | ActionHover | ActionInputText | ActionUrlChange | ActionWait

ActionBlockable = ActionClick | ActionHover | ActionInputText


class ActionTarget(BaseModel):
    class_name: str | None = None
    id: str | None = None
    mouse: "Mouse"
    sky_id: str | None = None
    tag_name: str | None = None
    texts: list[str] = []


class Mouse(BaseModel):
    xp: float | None = None
    """
    0 to 1.0 inclusive, percentage across the viewport
    """
    yp: float | None = None
    """
    0 to 1.0 inclusive, percentage down the viewport
    """


OutputBlock = t.Union[
    WorkflowDefinitionYamlBlocksItem_Action,
    WorkflowDefinitionYamlBlocksItem_GotoUrl,
    WorkflowDefinitionYamlBlocksItem_Wait,
]


class TargetInfo(BaseModel):
    attached: bool | None = None
    browserContextId: str | None = None
    canAccessOpener: bool | None = None
    targetId: str | None = None
    title: str | None = None
    type: str | None = None
    url: str | None = None


class CdpEventFrame(BaseModel):
    url: str | None = None

    class Config:
        extra = "allow"


class ExfiltratedEventCdpParams(BaseModel):
    # target_info_changed events
    targetInfo: TargetInfo | None = None

    # frame_requested_navigation events
    disposition: str | None = None
    frameId: str | None = None
    reason: str | None = None
    url: str | None = None

    # frame_navigated events
    frame: CdpEventFrame | None = None


class EventTarget(BaseModel):
    className: str | None = None
    id: str | None = None
    isHtml: bool = False
    isSvg: bool = False
    innerText: str | None = None
    skyId: str | None = None
    tagName: str | None = None
    text: list[str] = []
    value: str | int | None = None


class MousePosition(BaseModel):
    xa: float | None = None
    ya: float | None = None
    xp: float | None = None
    yp: float | None = None


class BoundingRect(BaseModel):
    bottom: float
    height: float
    left: float
    right: float
    top: float
    width: float
    x: float
    y: float


class Scroll(BaseModel):
    clientHeight: float
    clientWidth: float
    scrollHeight: float
    scrollLeft: float
    scrollTop: float
    scrollWidth: float


class ActiveElement(BaseModel):
    boundingRect: BoundingRect | None = None
    className: str | None = None
    id: str | None = None
    scroll: Scroll | None = None
    tagName: str | None = None


class Window(BaseModel):
    height: float
    scrollX: float
    scrollY: float
    width: float


class ExfiltratedEventConsoleParams(BaseModel):
    activeElement: ActiveElement
    code: str | None = None
    inputValue: str | None = None
    key: str | None = None
    mousePosition: MousePosition
    target: EventTarget
    timestamp: float
    type: str
    url: str
    window: Window


class ExfiltratedCdpEvent(BaseModel):
    kind: Literal["exfiltrated-event"]
    event_name: str
    params: ExfiltratedEventCdpParams
    source: Literal["cdp"]
    timestamp: float


class ExfiltratedConsoleEvent(BaseModel):
    kind: Literal["exfiltrated-event"]
    event_name: str
    params: ExfiltratedEventConsoleParams
    source: Literal["console"]
    timestamp: float


ExfiltratedEvent = ExfiltratedCdpEvent | ExfiltratedConsoleEvent


class StateMachineProtocol(t.Protocol):
    state: str

    def tick(self, event: ExfiltratedEvent, current_actions: list[Action]) -> Action | None: ...

    def on_action(self, action: Action, current_actions: list[Action]) -> bool: ...

    def reset(self) -> None: ...


# -- guards, predicates, etc.


def target_has_changed(current_target: EventTarget | None, event_target: EventTarget) -> bool:
    if not current_target:
        return False

    if not event_target.skyId and not event_target.id:
        return True  # sic: we cannot compare, so assume changed

    if not current_target.skyId and not current_target.id:
        return True  # sic: we cannot compare, so assume changed

    if current_target.id and event_target.id:
        if current_target.id != event_target.id:
            return True

    if current_target.skyId and event_target.skyId:
        if current_target.skyId != event_target.skyId:
            return True

    return False



import abc
from dataclasses import dataclass
from enum import StrEnum
from typing import Annotated, Any, Literal

import structlog
from pydantic import BaseModel, Field, field_validator, model_validator

from skyvern.config import settings
from skyvern.forge.sdk.workflow.models.parameter import OutputParameter, ParameterType, WorkflowParameterType
from skyvern.schemas.runs import GeoTarget, ProxyLocation, RunEngine
from skyvern.utils.strings import sanitize_identifier
from skyvern.utils.templating import replace_jinja_reference

LOG = structlog.get_logger()


def sanitize_block_label(value: str) -> str:
    """Sanitizes a block label to be a valid Python/Jinja2 identifier.

    Block labels are used to create output parameter keys (e.g., '{label}_output')
    which are then used as Jinja2 template variable names.

    Args:
        value: The raw label value to sanitize

    Returns:
        A sanitized label that is a valid Python identifier
    """
    return sanitize_identifier(value, default="block")


def sanitize_parameter_key(value: str) -> str:
    """Sanitizes a parameter key to be a valid Python/Jinja2 identifier.

    Parameter keys are used as Jinja2 template variable names.

    Args:
        value: The raw key value to sanitize

    Returns:
        A sanitized key that is a valid Python identifier
    """
    return sanitize_identifier(value, default="parameter")


def _replace_references_in_value(value: Any, old_key: str, new_key: str) -> Any:
    """Recursively replaces Jinja references in a value (string, dict, or list)."""
    if isinstance(value, str):
        return replace_jinja_reference(value, old_key, new_key)
    elif isinstance(value, dict):
        return {k: _replace_references_in_value(v, old_key, new_key) for k, v in value.items()}
    elif isinstance(value, list):
        return [_replace_references_in_value(item, old_key, new_key) for item in value]
    return value


def _replace_direct_string_in_value(value: Any, old_key: str, new_key: str) -> Any:
    """Recursively replaces exact string matches in a value (for fields like source_parameter_key)."""
    if isinstance(value, str):
        return new_key if value == old_key else value
    elif isinstance(value, dict):
        return {k: _replace_direct_string_in_value(v, old_key, new_key) for k, v in value.items()}
    elif isinstance(value, list):
        return [_replace_direct_string_in_value(item, old_key, new_key) for item in value]
    return value


def _make_unique(candidate: str, seen: set[str]) -> str:
    """Appends a numeric suffix to make candidate unique within the seen set.

    Args:
        candidate: The candidate identifier
        seen: Set of already-used identifiers (mutated -- the chosen name is added)

    Returns:
        A unique identifier, either candidate itself or candidate_N
    """
    if candidate not in seen:
        seen.add(candidate)
        return candidate
    counter = 2
    while f"{candidate}_{counter}" in seen:
        counter += 1
    unique = f"{candidate}_{counter}"
    seen.add(unique)
    return unique


def _sanitize_blocks_recursive(
    blocks: list[dict[str, Any]],
    label_mapping: dict[str, str],
    seen_labels: set[str],
) -> list[dict[str, Any]]:
    """Recursively sanitizes block labels and collects the label mapping.

    Args:
        blocks: List of block dictionaries
        label_mapping: Dictionary to store old_label -> new_label mappings (mutated)
        seen_labels: Set of already-used labels for collision avoidance (mutated)

    Returns:
        List of blocks with sanitized labels
    """
    sanitized_blocks = []
    for block in blocks:
        block = dict(block)  # Make a copy to avoid mutating the original

        # Sanitize the block's label
        if "label" in block and isinstance(block["label"], str):
            old_label = block["label"]
            new_label = sanitize_block_label(old_label)
            new_label = _make_unique(new_label, seen_labels)
            if old_label != new_label:
                label_mapping[old_label] = new_label
                block["label"] = new_label
            else:
                # Even if unchanged, track it as seen for collision avoidance
                seen_labels.add(old_label)

        # Handle nested blocks in for_loop
        if "loop_blocks" in block and isinstance(block["loop_blocks"], list):
            block["loop_blocks"] = _sanitize_blocks_recursive(block["loop_blocks"], label_mapping, seen_labels)

        sanitized_blocks.append(block)

    return sanitized_blocks


def _sanitize_parameters(
    parameters: list[dict[str, Any]],
    key_mapping: dict[str, str],
    seen_keys: set[str],
) -> list[dict[str, Any]]:
    """Sanitizes parameter keys and collects the key mapping.

    Args:
        parameters: List of parameter dictionaries
        key_mapping: Dictionary to store old_key -> new_key mappings (mutated)
        seen_keys: Set of already-used keys for collision avoidance (mutated)

    Returns:
        List of parameters with sanitized keys
    """
    sanitized_params = []
    for param in parameters:
        param = dict(param)  # Make a copy

        if "key" in param and isinstance(param["key"], str):
            old_key = param["key"]
            new_key = sanitize_parameter_key(old_key)
            new_key = _make_unique(new_key, seen_keys)
            if old_key != new_key:
                key_mapping[old_key] = new_key
                param["key"] = new_key
            else:
                # Even if unchanged, track it as seen for collision avoidance
                seen_keys.add(old_key)

        sanitized_params.append(param)

    return sanitized_params


def _update_parameter_keys_in_blocks(
    blocks: list[dict[str, Any]],
    key_mapping: dict[str, str],
) -> list[dict[str, Any]]:
    """Updates parameter_keys arrays in blocks to use new parameter key names.

    Args:
        blocks: List of block dictionaries
        key_mapping: Dictionary of old_key -> new_key mappings

    Returns:
        List of blocks with updated parameter_keys
    """
    updated_blocks = []
    for block in blocks:
        block = dict(block)

        # Update parameter_keys array if present
        if "parameter_keys" in block and isinstance(block["parameter_keys"], list):
            block["parameter_keys"] = [
                key_mapping.get(key, key) if isinstance(key, str) else key for key in block["parameter_keys"]
            ]

        # Handle nested blocks in for_loop
        if "loop_blocks" in block and isinstance(block["loop_blocks"], list):
            block["loop_blocks"] = _update_parameter_keys_in_blocks(block["loop_blocks"], key_mapping)

        updated_blocks.append(block)

    return updated_blocks


def sanitize_workflow_yaml_with_references(workflow_yaml: dict[str, Any]) -> dict[str, Any]:
    """Sanitizes block labels and parameter keys, and updates all references throughout the workflow.

    This function:
    1. Sanitizes all block labels to be valid Python identifiers
    2. Sanitizes all parameter keys to be valid Python identifiers
    3. Updates all Jinja references from {old_key} to {new_key}
    4. Updates next_block_label if it references an old label
    5. Updates finally_block_label if it references an old label
    6. Updates parameter_keys arrays in blocks
    7. Updates source_parameter_key and other direct references

    Args:
        workflow_yaml: The parsed workflow YAML dictionary

    Returns:
        The workflow YAML with sanitized identifiers and updated references
    """
    workflow_yaml = dict(workflow_yaml)  # Make a copy

    workflow_definition = workflow_yaml.get("workflow_definition")
    if not workflow_definition or not isinstance(workflow_definition, dict):
        return workflow_yaml

    workflow_definition = dict(workflow_definition)  # Make a copy
    workflow_yaml["workflow_definition"] = workflow_definition

    # Step 1: Sanitize all block labels and collect the mapping
    label_mapping: dict[str, str] = {}  # old_label -> new_label
    seen_labels: set[str] = set()
    blocks = workflow_definition.get("blocks")
    if blocks and isinstance(blocks, list):
        workflow_definition["blocks"] = _sanitize_blocks_recursive(blocks, label_mapping, seen_labels)

    # Step 2: Sanitize all parameter keys and collect the mapping
    param_key_mapping: dict[str, str] = {}  # old_key -> new_key
    seen_keys: set[str] = set()
    parameters = workflow_definition.get("parameters")
    if parameters and isinstance(parameters, list):
        workflow_definition["parameters"] = _sanitize_parameters(parameters, param_key_mapping, seen_keys)

    # If nothing was changed, return early
    if not label_mapping and not param_key_mapping:
        return workflow_yaml

    LOG.info(
        "Auto-sanitized workflow identifiers during import",
        sanitized_labels=label_mapping if label_mapping else None,
        sanitized_parameter_keys=param_key_mapping if param_key_mapping else None,
    )

    # Step 3: Update all block label references
    for old_label, new_label in label_mapping.items():
        old_output_key = f"{old_label}_output"
        new_output_key = f"{new_label}_output"

        # Update Jinja references in blocks for {label}_output pattern
        if "blocks" in workflow_definition:
            workflow_definition["blocks"] = _replace_references_in_value(
                workflow_definition["blocks"], old_output_key, new_output_key
            )
            # Also update shorthand {{ label }} references (must be done after _output to avoid partial matches)
            workflow_definition["blocks"] = _replace_references_in_value(
                workflow_definition["blocks"], old_label, new_label
            )

        # Update Jinja references in parameters for {label}_output pattern
        if "parameters" in workflow_definition:
            workflow_definition["parameters"] = _replace_references_in_value(
                workflow_definition["parameters"], old_output_key, new_output_key
            )
            # Also update shorthand {{ label }} references
            workflow_definition["parameters"] = _replace_references_in_value(
                workflow_definition["parameters"], old_label, new_label
            )
            # Also update direct string references (e.g., source_parameter_key)
            workflow_definition["parameters"] = _replace_direct_string_in_value(
                workflow_definition["parameters"], old_output_key, new_output_key
            )

    # Step 4: Update all parameter key references
    for old_key, new_key in param_key_mapping.items():
        # Update Jinja references in blocks (e.g., {{ old_key }})
        if "blocks" in workflow_definition:
            workflow_definition["blocks"] = _replace_references_in_value(
                workflow_definition["blocks"], old_key, new_key
            )

        # Update Jinja references in parameters (e.g., default values that reference other params)
        if "parameters" in workflow_definition:
            workflow_definition["parameters"] = _replace_references_in_value(
                workflow_definition["parameters"], old_key, new_key
            )
            # Also update direct string references (e.g., source_parameter_key)
            workflow_definition["parameters"] = _replace_direct_string_in_value(
                workflow_definition["parameters"], old_key, new_key
            )

    # Step 5: Update parameter_keys arrays in blocks
    if param_key_mapping and "blocks" in workflow_definition:
        workflow_definition["blocks"] = _update_parameter_keys_in_blocks(
            workflow_definition["blocks"], param_key_mapping
        )

    # Step 6: Update finally_block_label if it references an old label
    if "finally_block_label" in workflow_definition:
        finally_label = workflow_definition["finally_block_label"]
        if finally_label in label_mapping:
            workflow_definition["finally_block_label"] = label_mapping[finally_label]

    # Step 7: Update next_block_label in all blocks
    def update_next_block_label(blocks: list[dict[str, Any]]) -> list[dict[str, Any]]:
        updated_blocks = []
        for block in blocks:
            block = dict(block)
            if "next_block_label" in block:
                next_label = block["next_block_label"]
                if next_label in label_mapping:
                    block["next_block_label"] = label_mapping[next_label]
            if "loop_blocks" in block and isinstance(block["loop_blocks"], list):
                block["loop_blocks"] = update_next_block_label(block["loop_blocks"])
            updated_blocks.append(block)
        return updated_blocks

    if label_mapping and "blocks" in workflow_definition:
        workflow_definition["blocks"] = update_next_block_label(workflow_definition["blocks"])

    return workflow_yaml


class WorkflowStatus(StrEnum):
    published = "published"
    draft = "draft"
    auto_generated = "auto_generated"
    importing = "importing"
    import_failed = "import_failed"


class BlockType(StrEnum):
    TASK = "task"
    TaskV2 = "task_v2"
    FOR_LOOP = "for_loop"
    CONDITIONAL = "conditional"
    CODE = "code"
    TEXT_PROMPT = "text_prompt"
    DOWNLOAD_TO_S3 = "download_to_s3"
    UPLOAD_TO_S3 = "upload_to_s3"
    FILE_UPLOAD = "file_upload"
    SEND_EMAIL = "send_email"
    FILE_URL_PARSER = "file_url_parser"
    VALIDATION = "validation"
    ACTION = "action"
    NAVIGATION = "navigation"
    EXTRACTION = "extraction"
    LOGIN = "login"
    WAIT = "wait"
    FILE_DOWNLOAD = "file_download"
    GOTO_URL = "goto_url"
    PDF_PARSER = "pdf_parser"
    HTTP_REQUEST = "http_request"
    HUMAN_INTERACTION = "human_interaction"
    PRINT_PAGE = "print_page"


class BlockStatus(StrEnum):
    running = "running"
    completed = "completed"
    failed = "failed"
    terminated = "terminated"
    canceled = "canceled"
    timed_out = "timed_out"
    skipped = "skipped"


@dataclass(frozen=True)
class BlockResult:
    success: bool
    output_parameter: OutputParameter
    output_parameter_value: dict[str, Any] | list | str | None = None
    status: BlockStatus | None = None
    failure_reason: str | None = None
    workflow_run_block_id: str | None = None


class FileType(StrEnum):
    CSV = "csv"
    EXCEL = "excel"
    PDF = "pdf"
    IMAGE = "image"
    DOCX = "docx"


class PDFFormat(StrEnum):
    A4 = "A4"
    LETTER = "Letter"
    LEGAL = "Legal"
    TABLOID = "Tabloid"


class FileStorageType(StrEnum):
    S3 = "s3"
    AZURE = "azure"


class ParameterYAML(BaseModel, abc.ABC):
    parameter_type: ParameterType
    key: str
    description: str | None = None

    @field_validator("key")
    @classmethod
    def validate_key_is_valid_identifier(cls, v: str) -> str:
        """Validate that parameter key is a valid Jinja2/Python identifier.

        Parameter keys are used as Jinja2 template variable names. Jinja2 variable names
        must be valid Python identifiers (letters, digits, underscores; cannot start with digit).

        Characters like '/', '-', '.', etc. are NOT allowed because they are interpreted as
        operators in Jinja2 templates, causing parsing errors like "'State_' is undefined"
        when using a key like "State_/_Province".
        """
        if any(char in v for char in [" ", "\t", "\n", "\r"]):
            raise ValueError("Key cannot contain whitespace characters")

        if not v.isidentifier():
            raise ValueError(
                f"Key '{v}' is not a valid parameter name. "
                "Parameter keys must be valid Python identifiers "
                "(only letters, digits, and underscores; cannot start with a digit). "
                "Characters like '/', '-', '.', etc. are not allowed because they conflict with Jinja2 template syntax."
            )
        return v


class AWSSecretParameterYAML(ParameterYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the ParameterType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    parameter_type: Literal[ParameterType.AWS_SECRET] = ParameterType.AWS_SECRET  # type: ignore
    aws_key: str


class BitwardenLoginCredentialParameterYAML(ParameterYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the ParameterType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    parameter_type: Literal[ParameterType.BITWARDEN_LOGIN_CREDENTIAL] = ParameterType.BITWARDEN_LOGIN_CREDENTIAL  # type: ignore

    # bitwarden cli required fields
    bitwarden_client_id_aws_secret_key: str
    bitwarden_client_secret_aws_secret_key: str
    bitwarden_master_password_aws_secret_key: str
    # parameter key for the url to request the login credentials from bitwarden
    url_parameter_key: str | None = None
    # bitwarden collection id to filter the login credentials from,
    # if not provided, no filtering will be done
    bitwarden_collection_id: str | None = None
    # bitwarden item id to request the login credential
    bitwarden_item_id: str | None = None


class CredentialParameterYAML(ParameterYAML):
    parameter_type: Literal[ParameterType.CREDENTIAL] = ParameterType.CREDENTIAL  # type: ignore
    credential_id: str


class BitwardenSensitiveInformationParameterYAML(ParameterYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the ParameterType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    parameter_type: Literal["bitwarden_sensitive_information"] = ParameterType.BITWARDEN_SENSITIVE_INFORMATION  # type: ignore

    # bitwarden cli required fields
    bitwarden_client_id_aws_secret_key: str
    bitwarden_client_secret_aws_secret_key: str
    bitwarden_master_password_aws_secret_key: str
    # bitwarden collection id to filter the Bitwarden Identity from
    bitwarden_collection_id: str
    # unique key to identify the Bitwarden Identity in the collection
    # this has to be in the identity's name
    bitwarden_identity_key: str
    # fields to extract from the Bitwarden Identity. Custom fields are prioritized over default identity fields
    bitwarden_identity_fields: list[str]


class BitwardenCreditCardDataParameterYAML(ParameterYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the ParameterType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    parameter_type: Literal[ParameterType.BITWARDEN_CREDIT_CARD_DATA] = ParameterType.BITWARDEN_CREDIT_CARD_DATA  # type: ignore
    # bitwarden cli required fields
    bitwarden_client_id_aws_secret_key: str
    bitwarden_client_secret_aws_secret_key: str
    bitwarden_master_password_aws_secret_key: str
    # bitwarden ids for the credit card item
    bitwarden_collection_id: str
    bitwarden_item_id: str


class OnePasswordCredentialParameterYAML(ParameterYAML):
    parameter_type: Literal[ParameterType.ONEPASSWORD] = ParameterType.ONEPASSWORD  # type: ignore
    vault_id: str
    item_id: str


class AzureVaultCredentialParameterYAML(ParameterYAML):
    parameter_type: Literal[ParameterType.AZURE_VAULT_CREDENTIAL] = ParameterType.AZURE_VAULT_CREDENTIAL  # type: ignore
    vault_name: str
    username_key: str
    password_key: str
    totp_secret_key: str | None = None


class WorkflowParameterYAML(ParameterYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the ParameterType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    parameter_type: Literal[ParameterType.WORKFLOW] = ParameterType.WORKFLOW  # type: ignore
    workflow_parameter_type: WorkflowParameterType
    default_value: str | int | float | bool | dict | list | None = None


class ContextParameterYAML(ParameterYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the ParameterType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    parameter_type: Literal[ParameterType.CONTEXT] = ParameterType.CONTEXT  # type: ignore
    source_parameter_key: str


class OutputParameterYAML(ParameterYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the ParameterType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    parameter_type: Literal[ParameterType.OUTPUT] = ParameterType.OUTPUT  # type: ignore


class BlockYAML(BaseModel, abc.ABC):
    block_type: BlockType
    label: str = Field(description="Author-facing identifier; must be unique per workflow.")
    next_block_label: str | None = Field(
        default=None,
        description="Optional pointer to the label of the next block. "
        "When omitted, it will default to sequential order. See [[s-4bnl]].",
    )
    continue_on_failure: bool = False
    model: dict[str, Any] | None = None
    # Only valid for blocks inside a for loop block
    # Whether to continue to the next iteration when the block fails
    next_loop_on_failure: bool = False

    @field_validator("label")
    @classmethod
    def validate_label(cls, value: str) -> str:
        """Validate that block label is a valid Python identifier.

        Block labels are used to create output parameter keys (e.g., '{label}_output')
        which are then used as Jinja2 template variable names. Therefore, block labels
        must be valid Python identifiers.
        """
        if not value or not value.strip():
            raise ValueError("Block labels cannot be empty.")
        if not value.isidentifier():
            raise ValueError(
                f"Block label '{value}' is not a valid label. "
                "Block labels must be valid Python identifiers "
                "(only letters, digits, and underscores; cannot start with a digit). "
                "Characters like '/', '-', '.', etc. are not allowed because they conflict with Jinja2 template syntax."
            )
        return value


class TaskBlockYAML(BlockYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the BlockType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    block_type: Literal[BlockType.TASK] = BlockType.TASK  # type: ignore

    url: str | None = None
    title: str = ""
    engine: RunEngine = RunEngine.skyvern_v1
    navigation_goal: str | None = None
    data_extraction_goal: str | None = None
    data_schema: dict[str, Any] | list | str | None = None
    error_code_mapping: dict[str, str] | None = None
    max_retries: int = 0
    max_steps_per_run: int | None = None
    parameter_keys: list[str] | None = None
    complete_on_download: bool = False
    download_suffix: str | None = (
        None  # DEPRECATED: This field now sets the complete filename instead of appending to a random name
    )
    totp_verification_url: str | None = None
    totp_identifier: str | None = None
    disable_cache: bool = False
    complete_criterion: str | None = None
    terminate_criterion: str | None = None
    complete_verification: bool = True
    include_action_history_in_verification: bool = False


class ForLoopBlockYAML(BlockYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the BlockType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    block_type: Literal[BlockType.FOR_LOOP] = BlockType.FOR_LOOP  # type: ignore

    loop_blocks: list["BLOCK_YAML_SUBCLASSES"]
    loop_over_parameter_key: str = ""
    loop_variable_reference: str | None = None
    complete_if_empty: bool = False
    data_schema: dict[str, Any] | str | None = None


class BranchCriteriaYAML(BaseModel):
    criteria_type: Literal["jinja2_template", "prompt"] = "jinja2_template"
    expression: str
    description: str | None = None


class BranchConditionYAML(BaseModel):
    criteria: BranchCriteriaYAML | None = None
    next_block_label: str | None = None
    description: str | None = None
    is_default: bool = False

    @model_validator(mode="after")
    def validate_condition(cls, condition: "BranchConditionYAML") -> "BranchConditionYAML":
        if condition.criteria is None and not condition.is_default:
            raise ValueError("Branches without criteria must be marked as default.")
        if condition.criteria is not None and condition.is_default:
            raise ValueError("Default branches may not define criteria.")
        return condition


class ConditionalBlockYAML(BlockYAML):
    block_type: Literal[BlockType.CONDITIONAL] = BlockType.CONDITIONAL  # type: ignore

    branch_conditions: list[BranchConditionYAML] = Field(default_factory=list)

    @model_validator(mode="after")
    def validate_branches(cls, block: "ConditionalBlockYAML") -> "ConditionalBlockYAML":
        if not block.branch_conditions:
            raise ValueError("Conditional blocks require at least one branch.")

        default_branches = [branch for branch in block.branch_conditions if branch.is_default]
        if len(default_branches) > 1:
            raise ValueError("Only one default branch is permitted per conditional block.")

        return block


class CodeBlockYAML(BlockYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the BlockType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    block_type: Literal[BlockType.CODE] = BlockType.CODE  # type: ignore

    code: str
    parameter_keys: list[str] | None = None


class TextPromptBlockYAML(BlockYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the BlockType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    block_type: Literal[BlockType.TEXT_PROMPT] = BlockType.TEXT_PROMPT  # type: ignore

    llm_key: str | None = None
    prompt: str
    parameter_keys: list[str] | None = None
    json_schema: dict[str, Any] | None = None


class DownloadToS3BlockYAML(BlockYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the BlockType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    block_type: Literal[BlockType.DOWNLOAD_TO_S3] = BlockType.DOWNLOAD_TO_S3  # type: ignore

    url: str


class UploadToS3BlockYAML(BlockYAML):
    block_type: Literal[BlockType.UPLOAD_TO_S3] = BlockType.UPLOAD_TO_S3  # type: ignore

    path: str | None = None


class FileUploadBlockYAML(BlockYAML):
    block_type: Literal[BlockType.FILE_UPLOAD] = BlockType.FILE_UPLOAD  # type: ignore

    storage_type: FileStorageType = FileStorageType.S3
    s3_bucket: str | None = None
    aws_access_key_id: str | None = None
    aws_secret_access_key: str | None = None
    region_name: str | None = None
    azure_storage_account_name: str | None = None
    azure_storage_account_key: str | None = None
    azure_blob_container_name: str | None = None
    azure_folder_path: str | None = None
    path: str | None = None


class SendEmailBlockYAML(BlockYAML):
    # There is a mypy bug with Literal. Without the type: ignore, mypy will raise an error:
    # Parameter 1 of Literal[...] cannot be of type "Any"
    # This pattern already works in block.py but since the BlockType is not defined in this file, mypy is not able
    # to infer the type of the parameter_type attribute.
    block_type: Literal[BlockType.SEND_EMAIL] = BlockType.SEND_EMAIL  # type: ignore

    smtp_host_secret_parameter_key: str
    smtp_port_secret_parameter_key: str
    smtp_username_secret_parameter_key: str
    smtp_password_secret_parameter_key: str
    sender: str
    recipients: list[str]
    subject: str
    body: str
    file_attachments: list[str] | None = None


class FileParserBlockYAML(BlockYAML):
    block_type: Literal[BlockType.FILE_URL_PARSER] = BlockType.FILE_URL_PARSER  # type: ignore

    file_url: str
    file_type: FileType
    json_schema: dict[str, Any] | None = None


class PDFParserBlockYAML(BlockYAML):
    block_type: Literal[BlockType.PDF_PARSER] = BlockType.PDF_PARSER  # type: ignore

    file_url: str
    json_schema: dict[str, Any] | None = None


class ValidationBlockYAML(BlockYAML):
    block_type: Literal[BlockType.VALIDATION] = BlockType.VALIDATION  # type: ignore

    complete_criterion: str | None = None
    terminate_criterion: str | None = None
    error_code_mapping: dict[str, str] | None = None
    parameter_keys: list[str] | None = None
    disable_cache: bool = False


class ActionBlockYAML(BlockYAML):
    block_type: Literal[BlockType.ACTION] = BlockType.ACTION  # type: ignore

    url: str | None = None
    title: str = ""
    engine: RunEngine = RunEngine.skyvern_v1
    navigation_goal: str | None = None
    error_code_mapping: dict[str, str] | None = None
    max_retries: int = 0
    parameter_keys: list[str] | None = None
    complete_on_download: bool = False
    download_suffix: str | None = (
        None  # DEPRECATED: This field now sets the complete filename instead of appending to a random name
    )
    totp_verification_url: str | None = None
    totp_identifier: str | None = None
    disable_cache: bool = False


class NavigationBlockYAML(BlockYAML):
    block_type: Literal[BlockType.NAVIGATION] = BlockType.NAVIGATION  # type: ignore

    navigation_goal: str
    url: str | None = None
    title: str = ""
    engine: RunEngine = RunEngine.skyvern_v1
    error_code_mapping: dict[str, str] | None = None
    max_retries: int = 0
    max_steps_per_run: int | None = None
    parameter_keys: list[str] | None = None
    complete_on_download: bool = False
    download_suffix: str | None = (
        None  # DEPRECATED: This field now sets the complete filename instead of appending to a random name
    )
    totp_verification_url: str | None = None
    totp_identifier: str | None = None
    disable_cache: bool = False
    complete_criterion: str | None = None
    terminate_criterion: str | None = None
    complete_verification: bool = True
    include_action_history_in_verification: bool = False


class ExtractionBlockYAML(BlockYAML):
    block_type: Literal[BlockType.EXTRACTION] = BlockType.EXTRACTION  # type: ignore

    data_extraction_goal: str
    url: str | None = None
    title: str = ""
    engine: RunEngine = RunEngine.skyvern_v1
    data_schema: dict[str, Any] | list | str | None = None
    max_retries: int = 0
    max_steps_per_run: int | None = None
    parameter_keys: list[str] | None = None
    disable_cache: bool = False


class LoginBlockYAML(BlockYAML):
    block_type: Literal[BlockType.LOGIN] = BlockType.LOGIN  # type: ignore

    url: str | None = None
    title: str = ""
    engine: RunEngine = RunEngine.skyvern_v1
    navigation_goal: str | None = None
    error_code_mapping: dict[str, str] | None = None
    max_retries: int = 0
    max_steps_per_run: int | None = None
    parameter_keys: list[str] | None = None
    totp_verification_url: str | None = None
    totp_identifier: str | None = None
    disable_cache: bool = False
    complete_criterion: str | None = None
    terminate_criterion: str | None = None
    complete_verification: bool = True


class WaitBlockYAML(BlockYAML):
    block_type: Literal[BlockType.WAIT] = BlockType.WAIT  # type: ignore
    wait_sec: int = 0


class HumanInteractionBlockYAML(BlockYAML):
    block_type: Literal[BlockType.HUMAN_INTERACTION] = BlockType.HUMAN_INTERACTION  # type: ignore

    instructions: str = "Please review and approve or reject to continue the workflow."
    positive_descriptor: str = "Approve"
    negative_descriptor: str = "Reject"
    timeout_seconds: int

    sender: str
    recipients: list[str]
    subject: str
    body: str


class FileDownloadBlockYAML(BlockYAML):
    block_type: Literal[BlockType.FILE_DOWNLOAD] = BlockType.FILE_DOWNLOAD  # type: ignore

    navigation_goal: str
    url: str | None = None
    title: str = ""
    engine: RunEngine = RunEngine.skyvern_v1
    error_code_mapping: dict[str, str] | None = None
    max_retries: int = 0
    max_steps_per_run: int | None = None
    parameter_keys: list[str] | None = None
    download_suffix: str | None = (
        None  # DEPRECATED: This field now sets the complete filename instead of appending to a random name
    )
    totp_verification_url: str | None = None
    totp_identifier: str | None = None
    disable_cache: bool = False
    download_timeout: float | None = None


class UrlBlockYAML(BlockYAML):
    block_type: Literal[BlockType.GOTO_URL] = BlockType.GOTO_URL  # type: ignore
    url: str


class TaskV2BlockYAML(BlockYAML):
    block_type: Literal[BlockType.TaskV2] = BlockType.TaskV2  # type: ignore
    prompt: str
    url: str | None = None
    totp_verification_url: str | None = None
    totp_identifier: str | None = None
    max_iterations: int = settings.MAX_ITERATIONS_PER_TASK_V2
    max_steps: int = settings.MAX_STEPS_PER_TASK_V2
    disable_cache: bool = False


class HttpRequestBlockYAML(BlockYAML):
    block_type: Literal[BlockType.HTTP_REQUEST] = BlockType.HTTP_REQUEST  # type: ignore

    # Individual HTTP parameters
    method: str = "GET"
    url: str | None = None
    headers: dict[str, str] | None = None
    body: dict[str, Any] | None = None  # Changed to consistently be dict only
    files: dict[str, str] | None = None  # Dictionary mapping field names to file paths/URLs for multipart file uploads
    timeout: int = 30
    follow_redirects: bool = True
    download_filename: str | None = None
    save_response_as_file: bool = False

    # Parameter keys for templating
    parameter_keys: list[str] | None = None


class PrintPageBlockYAML(BlockYAML):
    block_type: Literal[BlockType.PRINT_PAGE] = BlockType.PRINT_PAGE  # type: ignore
    include_timestamp: bool = True
    custom_filename: str | None = None
    format: PDFFormat = PDFFormat.A4
    landscape: bool = False
    print_background: bool = True
    parameter_keys: list[str] | None = None


PARAMETER_YAML_SUBCLASSES = (
    AWSSecretParameterYAML
    | BitwardenLoginCredentialParameterYAML
    | BitwardenSensitiveInformationParameterYAML
    | BitwardenCreditCardDataParameterYAML
    | OnePasswordCredentialParameterYAML
    | AzureVaultCredentialParameterYAML
    | WorkflowParameterYAML
    | ContextParameterYAML
    | OutputParameterYAML
    | CredentialParameterYAML
)
PARAMETER_YAML_TYPES = Annotated[PARAMETER_YAML_SUBCLASSES, Field(discriminator="parameter_type")]

BLOCK_YAML_SUBCLASSES = (
    TaskBlockYAML
    | ForLoopBlockYAML
    | CodeBlockYAML
    | TextPromptBlockYAML
    | DownloadToS3BlockYAML
    | UploadToS3BlockYAML
    | FileUploadBlockYAML
    | SendEmailBlockYAML
    | FileParserBlockYAML
    | ValidationBlockYAML
    | ActionBlockYAML
    | NavigationBlockYAML
    | ExtractionBlockYAML
    | LoginBlockYAML
    | WaitBlockYAML
    | HumanInteractionBlockYAML
    | FileDownloadBlockYAML
    | UrlBlockYAML
    | PDFParserBlockYAML
    | TaskV2BlockYAML
    | HttpRequestBlockYAML
    | ConditionalBlockYAML
    | PrintPageBlockYAML
)
BLOCK_YAML_TYPES = Annotated[BLOCK_YAML_SUBCLASSES, Field(discriminator="block_type")]


class WorkflowDefinitionYAML(BaseModel):
    version: int = 1
    parameters: list[PARAMETER_YAML_TYPES]
    blocks: list[BLOCK_YAML_TYPES]
    finally_block_label: str | None = None

    @model_validator(mode="after")
    def validate_unique_block_labels(cls, workflow: "WorkflowDefinitionYAML") -> "WorkflowDefinitionYAML":
        labels = [block.label for block in workflow.blocks]
        duplicates = [label for label in labels if labels.count(label) > 1]

        if duplicates:
            unique_duplicates = sorted(set(duplicates))
            raise ValueError(
                f"Block labels must be unique within a workflow. "
                f"Found duplicate label(s): {', '.join(unique_duplicates)}"
            )

        if workflow.finally_block_label and workflow.finally_block_label not in labels:
            raise ValueError(
                f"finally_block_label '{workflow.finally_block_label}' does not reference a valid block. "
                f"Available labels: {', '.join(labels) if labels else '(none)'}"
            )

        return workflow


class WorkflowCreateYAMLRequest(BaseModel):
    title: str
    description: str | None = None
    proxy_location: ProxyLocation | GeoTarget | dict | None = None
    webhook_callback_url: str | None = None
    totp_verification_url: str | None = None
    totp_identifier: str | None = None
    persist_browser_session: bool = False
    model: dict[str, Any] | None = None
    workflow_definition: WorkflowDefinitionYAML
    is_saved_task: bool = False
    max_screenshot_scrolls: int | None = None
    extra_http_headers: dict[str, str] | None = None
    status: WorkflowStatus = WorkflowStatus.published
    run_with: str | None = None
    ai_fallback: bool = False
    cache_key: str | None = "default"
    run_sequentially: bool = False
    sequential_key: str | None = None
    folder_id: str | None = None


class WorkflowRequest(BaseModel):
    json_definition: WorkflowCreateYAMLRequest | None = Field(
        default=None,
        description="Workflow definition in JSON format",
    )
    yaml_definition: str | None = Field(
        default=None,
        description="Workflow definition in YAML format",
    )



import os
import shutil
import uuid
from datetime import datetime, timezone
from typing import BinaryIO

import structlog
import zstandard as zstd

from skyvern.config import settings
from skyvern.constants import BROWSER_DOWNLOADING_SUFFIX, DOWNLOAD_FILE_PREFIX
from skyvern.forge.sdk.api.aws import AsyncAWSClient, S3StorageClass
from skyvern.forge.sdk.api.files import (
    calculate_sha256_for_file,
    create_named_temporary_file,
    get_download_dir,
    get_skyvern_temp_dir,
    make_temp_directory,
    unzip_files,
)
from skyvern.forge.sdk.artifact.models import Artifact, ArtifactType, LogEntityType
from skyvern.forge.sdk.artifact.storage.base import FILE_EXTENTSION_MAP, BaseStorage
from skyvern.forge.sdk.models import Step
from skyvern.forge.sdk.schemas.ai_suggestions import AISuggestion
from skyvern.forge.sdk.schemas.files import FileInfo
from skyvern.forge.sdk.schemas.task_v2 import TaskV2, Thought
from skyvern.forge.sdk.schemas.workflow_runs import WorkflowRunBlock

LOG = structlog.get_logger()


S3_ZSTD_COMPRESSED_SUFFIX = ".zst"


class S3Storage(BaseStorage):
    _PATH_VERSION = "v1"

    def __init__(self, bucket: str | None = None, endpoint_url: str | None = None) -> None:
        self.async_client = AsyncAWSClient(endpoint_url=endpoint_url)
        self.bucket = bucket or settings.AWS_S3_BUCKET_ARTIFACTS

    def build_uri(self, *, organization_id: str, artifact_id: str, step: Step, artifact_type: ArtifactType) -> str:
        file_ext = FILE_EXTENTSION_MAP[artifact_type]
        if artifact_type == ArtifactType.HAR:
            file_ext = f"{file_ext}{S3_ZSTD_COMPRESSED_SUFFIX}"

        return f"{self._build_base_uri(organization_id)}/{step.task_id}/{step.order:02d}_{step.retry_index}_{step.step_id}/{datetime.utcnow().isoformat()}_{artifact_id}_{artifact_type}.{file_ext}"

    async def retrieve_global_workflows(self) -> list[str]:
        uri = f"s3://{self.bucket}/{settings.ENV}/global_workflows.txt"
        data = await self.async_client.download_file(uri, log_exception=False)
        if not data:
            return []
        return [line.strip() for line in data.decode("utf-8").split("\n") if line.strip()]

    def _build_base_uri(self, organization_id: str) -> str:
        return f"s3://{self.bucket}/{self._PATH_VERSION}/{settings.ENV}/{organization_id}"

    def build_log_uri(
        self, *, organization_id: str, log_entity_type: LogEntityType, log_entity_id: str, artifact_type: ArtifactType
    ) -> str:
        file_ext = FILE_EXTENTSION_MAP[artifact_type]
        return f"{self._build_base_uri(organization_id)}/logs/{log_entity_type}/{log_entity_id}/{datetime.utcnow().isoformat()}_{artifact_type}.{file_ext}"

    def build_thought_uri(
        self, *, organization_id: str, artifact_id: str, thought: Thought, artifact_type: ArtifactType
    ) -> str:
        file_ext = FILE_EXTENTSION_MAP[artifact_type]
        return f"{self._build_base_uri(organization_id)}/observers/{thought.observer_cruise_id}/{thought.observer_thought_id}/{datetime.utcnow().isoformat()}_{artifact_id}_{artifact_type}.{file_ext}"

    def build_task_v2_uri(
        self, *, organization_id: str, artifact_id: str, task_v2: TaskV2, artifact_type: ArtifactType
    ) -> str:
        file_ext = FILE_EXTENTSION_MAP[artifact_type]
        return f"{self._build_base_uri(organization_id)}/observers/{task_v2.observer_cruise_id}/{datetime.utcnow().isoformat()}_{artifact_id}_{artifact_type}.{file_ext}"

    def build_workflow_run_block_uri(
        self,
        *,
        organization_id: str,
        artifact_id: str,
        workflow_run_block: WorkflowRunBlock,
        artifact_type: ArtifactType,
    ) -> str:
        file_ext = FILE_EXTENTSION_MAP[artifact_type]
        return f"{self._build_base_uri(organization_id)}/workflow_runs/{workflow_run_block.workflow_run_id}/{workflow_run_block.workflow_run_block_id}/{datetime.utcnow().isoformat()}_{artifact_id}_{artifact_type}.{file_ext}"

    def build_ai_suggestion_uri(
        self, *, organization_id: str, artifact_id: str, ai_suggestion: AISuggestion, artifact_type: ArtifactType
    ) -> str:
        file_ext = FILE_EXTENTSION_MAP[artifact_type]
        return f"{self._build_base_uri(organization_id)}/ai_suggestions/{ai_suggestion.ai_suggestion_id}/{datetime.utcnow().isoformat()}_{artifact_id}_{artifact_type}.{file_ext}"

    def build_script_file_uri(
        self, *, organization_id: str, script_id: str, script_version: int, file_path: str
    ) -> str:
        """Build the S3 URI for a script file.

        Args:
            organization_id: The organization ID
            script_id: The script ID
            script_version: The script version
            file_path: The file path relative to script root

        Returns:
            The S3 URI for the script file
        """
        return f"{self._build_base_uri(organization_id)}/scripts/{script_id}/{script_version}/{file_path}"

    async def store_artifact(self, artifact: Artifact, data: bytes) -> None:
        # We compress HAR files with zstd level 3 to reduce storage size.
        # HARs are easily compressible because they are mostly JSON.
        # Other artifacts are not compressed because they are not easily compressible.
        uri = artifact.uri
        if uri.endswith(S3_ZSTD_COMPRESSED_SUFFIX):
            cctx = zstd.ZstdCompressor(level=3)
            data = cctx.compress(data)

        sc = await self._get_storage_class_for_org(artifact.organization_id, self.bucket, len(data))
        LOG.debug(
            "Storing artifact",
            artifact_id=artifact.artifact_id,
            organization_id=artifact.organization_id,
            uri=uri,
            storage_class=sc,
        )
        await self.async_client.upload_file(uri, data, storage_class=sc)

    async def _get_storage_class_for_org(
        self,
        organization_id: str,
        bucket: str,
        object_size_bytes: int | None = None,
    ) -> S3StorageClass:
        return S3StorageClass.STANDARD

    async def retrieve_artifact(self, artifact: Artifact) -> bytes | None:
        data = await self.async_client.download_file(artifact.uri)
        # Decompress zstd-compressed files
        if data and artifact.uri.endswith(S3_ZSTD_COMPRESSED_SUFFIX):
            dctx = zstd.ZstdDecompressor()
            data = dctx.decompress(data)
        return data

    async def get_share_link(self, artifact: Artifact) -> str | None:
        share_urls = await self.async_client.create_presigned_urls([artifact.uri])
        return share_urls[0] if share_urls else None

    async def get_share_links(self, artifacts: list[Artifact]) -> list[str] | None:
        return await self.async_client.create_presigned_urls([artifact.uri for artifact in artifacts])

    async def store_artifact_from_path(self, artifact: Artifact, path: str) -> None:
        sc = await self._get_storage_class_for_org(artifact.organization_id, self.bucket, os.path.getsize(path))
        LOG.debug(
            "Storing artifact from path",
            artifact_id=artifact.artifact_id,
            organization_id=artifact.organization_id,
            uri=artifact.uri,
            storage_class=sc,
            path=path,
        )
        await self.async_client.upload_file_from_path(artifact.uri, path, storage_class=sc)

    async def save_streaming_file(self, organization_id: str, file_name: str) -> None:
        from_path = f"{get_skyvern_temp_dir()}/{organization_id}/{file_name}"
        to_path = f"s3://{settings.AWS_S3_BUCKET_SCREENSHOTS}/{settings.ENV}/{organization_id}/{file_name}"
        sc = await self._get_storage_class_for_org(organization_id, settings.AWS_S3_BUCKET_SCREENSHOTS)
        LOG.debug(
            "Saving streaming file",
            organization_id=organization_id,
            file_name=file_name,
            from_path=from_path,
            to_path=to_path,
            storage_class=sc,
        )
        await self.async_client.upload_file_from_path(to_path, from_path, storage_class=sc)

    async def get_streaming_file(self, organization_id: str, file_name: str, use_default: bool = True) -> bytes | None:
        path = f"s3://{settings.AWS_S3_BUCKET_SCREENSHOTS}/{settings.ENV}/{organization_id}/{file_name}"
        return await self.async_client.download_file(path, log_exception=False)

    async def store_browser_session(self, organization_id: str, workflow_permanent_id: str, directory: str) -> None:
        # Zip the directory to a temp file
        temp_zip_file = create_named_temporary_file()
        zip_file_path = shutil.make_archive(temp_zip_file.name, "zip", directory)
        browser_session_uri = f"s3://{settings.AWS_S3_BUCKET_BROWSER_SESSIONS}/{settings.ENV}/{organization_id}/{workflow_permanent_id}.zip"
        sc = await self._get_storage_class_for_org(organization_id, settings.AWS_S3_BUCKET_BROWSER_SESSIONS)
        LOG.debug(
            "Storing browser session",
            organization_id=organization_id,
            workflow_permanent_id=workflow_permanent_id,
            zip_file_path=zip_file_path,
            browser_session_uri=browser_session_uri,
            storage_class=sc,
        )
        await self.async_client.upload_file_from_path(browser_session_uri, zip_file_path, storage_class=sc)

    async def retrieve_browser_session(self, organization_id: str, workflow_permanent_id: str) -> str | None:
        browser_session_uri = f"s3://{settings.AWS_S3_BUCKET_BROWSER_SESSIONS}/{settings.ENV}/{organization_id}/{workflow_permanent_id}.zip"
        downloaded_zip_bytes = await self.async_client.download_file(browser_session_uri, log_exception=True)
        if not downloaded_zip_bytes:
            return None
        temp_zip_file = create_named_temporary_file(delete=False)
        temp_zip_file.write(downloaded_zip_bytes)
        temp_zip_file_path = temp_zip_file.name

        temp_dir = make_temp_directory(prefix="skyvern_browser_session_")
        unzip_files(temp_zip_file_path, temp_dir)
        temp_zip_file.close()
        return temp_dir

    async def store_browser_profile(self, organization_id: str, profile_id: str, directory: str) -> None:
        """Store browser profile to S3."""
        temp_zip_file = create_named_temporary_file()
        zip_file_path = shutil.make_archive(temp_zip_file.name, "zip", directory)
        profile_uri = (
            f"s3://{settings.AWS_S3_BUCKET_BROWSER_SESSIONS}/{settings.ENV}/{organization_id}/profiles/{profile_id}.zip"
        )
        sc = await self._get_storage_class_for_org(organization_id, settings.AWS_S3_BUCKET_BROWSER_SESSIONS)
        LOG.debug(
            "Storing browser profile",
            organization_id=organization_id,
            profile_id=profile_id,
            zip_file_path=zip_file_path,
            profile_uri=profile_uri,
            storage_class=sc,
        )
        await self.async_client.upload_file_from_path(profile_uri, zip_file_path, storage_class=sc)

    async def retrieve_browser_profile(self, organization_id: str, profile_id: str) -> str | None:
        """Retrieve browser profile from S3."""
        profile_uri = (
            f"s3://{settings.AWS_S3_BUCKET_BROWSER_SESSIONS}/{settings.ENV}/{organization_id}/profiles/{profile_id}.zip"
        )
        downloaded_zip_bytes = await self.async_client.download_file(profile_uri, log_exception=True)
        if not downloaded_zip_bytes:
            return None
        temp_zip_file = create_named_temporary_file(delete=False)
        temp_zip_file.write(downloaded_zip_bytes)
        temp_zip_file_path = temp_zip_file.name

        temp_dir = make_temp_directory(prefix="skyvern_browser_profile_")
        unzip_files(temp_zip_file_path, temp_dir)
        temp_zip_file.close()
        return temp_dir

    async def list_downloaded_files_in_browser_session(
        self, organization_id: str, browser_session_id: str
    ) -> list[str]:
        bucket = settings.AWS_S3_BUCKET_ARTIFACTS
        uri = f"s3://{bucket}/{self._PATH_VERSION}/{settings.ENV}/{organization_id}/browser_sessions/{browser_session_id}/downloads"
        return [f"s3://{bucket}/{file}" for file in await self.async_client.list_files(uri=uri)]

    async def get_shared_downloaded_files_in_browser_session(
        self, organization_id: str, browser_session_id: str
    ) -> list[FileInfo]:
        object_keys = await self.list_downloaded_files_in_browser_session(organization_id, browser_session_id)
        if len(object_keys) == 0:
            return []

        file_infos: list[FileInfo] = []
        for key in object_keys:
            metadata = {}
            modified_at: datetime | None = None
            # Get metadata (including checksum)
            try:
                object_info = await self.async_client.get_object_info(key)
                metadata = object_info.get("Metadata", {})
                modified_at = object_info.get("LastModified")
            except Exception:
                LOG.exception("Object info retrieval failed", uri=key)

            # Create FileInfo object
            filename = os.path.basename(key)
            checksum = metadata.get("sha256_checksum") if metadata else None

            # Get presigned URL
            presigned_urls = await self.async_client.create_presigned_urls([key])
            if not presigned_urls:
                continue

            file_info = FileInfo(
                url=presigned_urls[0],
                checksum=checksum,
                filename=metadata.get("original_filename", filename) if metadata else filename,
                modified_at=modified_at,
            )
            file_infos.append(file_info)

        return file_infos

    async def list_downloading_files_in_browser_session(
        self, organization_id: str, browser_session_id: str
    ) -> list[str]:
        bucket = settings.AWS_S3_BUCKET_ARTIFACTS
        uri = f"s3://{bucket}/{self._PATH_VERSION}/{settings.ENV}/{organization_id}/browser_sessions/{browser_session_id}/downloads"
        files = [f"s3://{bucket}/{file}" for file in await self.async_client.list_files(uri=uri)]
        return [file for file in files if file.endswith(BROWSER_DOWNLOADING_SUFFIX)]

    async def list_recordings_in_browser_session(self, organization_id: str, browser_session_id: str) -> list[str]:
        """List all recording files for a browser session from S3."""
        bucket = settings.AWS_S3_BUCKET_ARTIFACTS
        uri = f"s3://{bucket}/{self._PATH_VERSION}/{settings.ENV}/{organization_id}/browser_sessions/{browser_session_id}/videos"
        return [f"s3://{bucket}/{file}" for file in await self.async_client.list_files(uri=uri)]

    async def get_shared_recordings_in_browser_session(
        self, organization_id: str, browser_session_id: str
    ) -> list[FileInfo]:
        """Get recording files with presigned URLs for a browser session."""
        object_keys = await self.list_recordings_in_browser_session(organization_id, browser_session_id)
        if len(object_keys) == 0:
            return []

        file_infos: list[FileInfo] = []
        for key in object_keys:
            # Playwright's record_video_dir should only contain .webm files.
            # Filter defensively in case of unexpected files.
            key_lower = key.lower()
            if not (key_lower.endswith(".webm") or key_lower.endswith(".mp4")):
                LOG.warning(
                    "Skipping recording file with unsupported extension",
                    uri=key,
                    organization_id=organization_id,
                    browser_session_id=browser_session_id,
                )
                continue

            metadata = {}
            modified_at: datetime | None = None
            content_length: int | None = None
            # Get metadata (including checksum)
            try:
                object_info = await self.async_client.get_object_info(key)
                metadata = object_info.get("Metadata", {})
                modified_at = object_info.get("LastModified")
                content_length = object_info.get("ContentLength")
            except Exception:
                LOG.exception("Recording object info retrieval failed", uri=key)

            # Skip zero-byte objects (if any incompleted uploads)
            if content_length == 0:
                continue

            # Create FileInfo object
            filename = os.path.basename(key)
            checksum = metadata.get("sha256_checksum") if metadata else None

            # Get presigned URL
            presigned_urls = await self.async_client.create_presigned_urls([key])
            if not presigned_urls:
                continue

            file_info = FileInfo(
                url=presigned_urls[0],
                checksum=checksum,
                filename=metadata.get("original_filename", filename) if metadata else filename,
                modified_at=modified_at,
            )
            file_infos.append(file_info)

        # Prefer the newest recording first (S3 list order is not guaranteed).
        # Treat None as "oldest".
        file_infos.sort(key=lambda f: (f.modified_at is not None, f.modified_at), reverse=True)
        return file_infos

    async def save_downloaded_files(self, organization_id: str, run_id: str | None) -> None:
        download_dir = get_download_dir(run_id=run_id)
        files = os.listdir(download_dir)
        sc = await self._get_storage_class_for_org(organization_id, settings.AWS_S3_BUCKET_UPLOADS)
        base_uri = (
            f"s3://{settings.AWS_S3_BUCKET_UPLOADS}/{DOWNLOAD_FILE_PREFIX}/{settings.ENV}/{organization_id}/{run_id}"
        )
        for file in files:
            fpath = os.path.join(download_dir, file)
            if not os.path.isfile(fpath):
                continue
            uri = f"{base_uri}/{file}"
            checksum = calculate_sha256_for_file(fpath)
            LOG.info(
                "Calculated checksum for file",
                file=file,
                checksum=checksum,
                organization_id=organization_id,
                storage_class=sc,
            )
            # Upload file with checksum metadata
            await self.async_client.upload_file_from_path(
                uri=uri,
                file_path=fpath,
                metadata={"sha256_checksum": checksum, "original_filename": file},
                storage_class=sc,
            )

    async def get_downloaded_files(self, organization_id: str, run_id: str | None) -> list[FileInfo]:
        bucket = settings.AWS_S3_BUCKET_UPLOADS
        uri = f"s3://{bucket}/{DOWNLOAD_FILE_PREFIX}/{settings.ENV}/{organization_id}/{run_id}"
        object_keys = await self.async_client.list_files(uri=uri)
        if len(object_keys) == 0:
            return []

        file_infos: list[FileInfo] = []
        for key in object_keys:
            object_uri = f"s3://{bucket}/{key}"

            # Get metadata (including checksum)
            metadata = await self.async_client.get_file_metadata(object_uri, log_exception=False)

            # Create FileInfo object
            filename = os.path.basename(key)
            checksum = metadata.get("sha256_checksum") if metadata else None

            # Get presigned URL
            presigned_urls = await self.async_client.create_presigned_urls([object_uri])
            if not presigned_urls:
                continue

            file_info = FileInfo(
                url=presigned_urls[0],
                checksum=checksum,
                filename=metadata.get("original_filename", filename) if metadata else filename,
            )
            file_infos.append(file_info)

        return file_infos

    async def save_legacy_file(
        self, *, organization_id: str, filename: str, fileObj: BinaryIO
    ) -> tuple[str, str] | None:
        todays_date = datetime.now(tz=timezone.utc).strftime("%Y-%m-%d")
        bucket = settings.AWS_S3_BUCKET_UPLOADS
        sc = await self._get_storage_class_for_org(organization_id, bucket)
        # First try uploading with original filename
        try:
            sanitized_filename = os.path.basename(filename)  # Remove any path components
            s3_uri = f"s3://{bucket}/{settings.ENV}/{organization_id}/{todays_date}/{sanitized_filename}"
            uploaded_s3_uri = await self.async_client.upload_file_stream(s3_uri, fileObj, storage_class=sc)
        except Exception:
            LOG.error("Failed to upload file to S3", exc_info=True)
            uploaded_s3_uri = None

        # If upload fails, try again with UUID prefix
        if not uploaded_s3_uri:
            uuid_prefixed_filename = f"{str(uuid.uuid4())}_{filename}"
            s3_uri = f"s3://{bucket}/{settings.ENV}/{organization_id}/{todays_date}/{uuid_prefixed_filename}"
            fileObj.seek(0)  # Reset file pointer
            uploaded_s3_uri = await self.async_client.upload_file_stream(s3_uri, fileObj, storage_class=sc)

        if not uploaded_s3_uri:
            LOG.error(
                "Failed to upload file to S3 after retrying with UUID prefix",
                organization_id=organization_id,
                storage_class=sc,
                filename=filename,
                exc_info=True,
            )
            return None
        LOG.debug(
            "Legacy file upload",
            organization_id=organization_id,
            storage_class=sc,
            filename=filename,
            uploaded_s3_uri=uploaded_s3_uri,
        )
        # Generate a presigned URL for the uploaded file
        presigned_urls = await self.async_client.create_presigned_urls([uploaded_s3_uri])
        if not presigned_urls:
            LOG.error(
                "Failed to create presigned URL for uploaded file",
                organization_id=organization_id,
                storage_class=sc,
                uploaded_s3_uri=uploaded_s3_uri,
                filename=filename,
                exc_info=True,
            )
            return None
        return presigned_urls[0], uploaded_s3_uri

    def _build_browser_session_uri(
        self,
        organization_id: str,
        browser_session_id: str,
        artifact_type: str,
        remote_path: str,
        date: str | None = None,
    ) -> str:
        """Build the S3 URI for a browser session file."""
        base = f"s3://{self.bucket}/{self._PATH_VERSION}/{settings.ENV}/{organization_id}/browser_sessions/{browser_session_id}/{artifact_type}"
        if date:
            return f"{base}/{date}/{remote_path}"
        return f"{base}/{remote_path}"

    async def sync_browser_session_file(
        self,
        organization_id: str,
        browser_session_id: str,
        artifact_type: str,
        local_file_path: str,
        remote_path: str,
        date: str | None = None,
    ) -> str:
        """Sync a file from local browser session to S3."""
        uri = self._build_browser_session_uri(organization_id, browser_session_id, artifact_type, remote_path, date)
        sc = await self._get_storage_class_for_org(organization_id, self.bucket)
        await self.async_client.upload_file_from_path(uri, local_file_path, storage_class=sc)
        return uri

    async def delete_browser_session_file(
        self,
        organization_id: str,
        browser_session_id: str,
        artifact_type: str,
        remote_path: str,
        date: str | None = None,
    ) -> None:
        """Delete a file from browser session storage in S3."""
        uri = self._build_browser_session_uri(organization_id, browser_session_id, artifact_type, remote_path, date)
        await self.async_client.delete_file(uri, log_exception=True)

    async def browser_session_file_exists(
        self,
        organization_id: str,
        browser_session_id: str,
        artifact_type: str,
        remote_path: str,
        date: str | None = None,
    ) -> bool:
        """Check if a file exists in browser session storage in S3."""
        uri = self._build_browser_session_uri(organization_id, browser_session_id, artifact_type, remote_path, date)
        try:
            info = await self.async_client.get_object_info(uri)
            return info is not None
        except Exception:
            return False

    async def download_uploaded_file(self, uri: str) -> bytes | None:
        """Download a user-uploaded file from S3."""
        return await self.async_client.download_file(uri, log_exception=False)

    async def file_exists(self, uri: str) -> bool:
        """Check if a file exists at the given S3 URI."""
        try:
            info = await self.async_client.get_object_info(uri)
            return info is not None
        except Exception:
            return False

    @property
    def storage_type(self) -> str:
        """Returns 's3' as the storage type."""
        return "s3"



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .act_action_data import ActActionData
from .click_action_data import ClickActionData
from .extract_action_data import ExtractActionData
from .extract_action_extract_schema import ExtractActionExtractSchema
from .input_text_action_data import InputTextActionData
from .select_option_action_data import SelectOptionActionData
from .upload_file_action_data import UploadFileActionData


class RunSdkActionRequestAction_AiAct(UniversalBaseModel):
    """
    The action to execute with its specific parameters
    """

    type: typing.Literal["ai_act"] = "ai_act"
    intention: typing.Optional[str] = None
    data: typing.Optional[ActActionData] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class RunSdkActionRequestAction_AiClick(UniversalBaseModel):
    """
    The action to execute with its specific parameters
    """

    type: typing.Literal["ai_click"] = "ai_click"
    selector: typing.Optional[str] = None
    intention: typing.Optional[str] = None
    data: typing.Optional[ClickActionData] = None
    timeout: typing.Optional[float] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class RunSdkActionRequestAction_AiInputText(UniversalBaseModel):
    """
    The action to execute with its specific parameters
    """

    type: typing.Literal["ai_input_text"] = "ai_input_text"
    selector: typing.Optional[str] = None
    value: typing.Optional[str] = None
    intention: typing.Optional[str] = None
    data: typing.Optional[InputTextActionData] = None
    totp_identifier: typing.Optional[str] = None
    totp_url: typing.Optional[str] = None
    timeout: typing.Optional[float] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class RunSdkActionRequestAction_AiSelectOption(UniversalBaseModel):
    """
    The action to execute with its specific parameters
    """

    type: typing.Literal["ai_select_option"] = "ai_select_option"
    selector: typing.Optional[str] = None
    value: typing.Optional[str] = None
    intention: typing.Optional[str] = None
    data: typing.Optional[SelectOptionActionData] = None
    timeout: typing.Optional[float] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class RunSdkActionRequestAction_AiUploadFile(UniversalBaseModel):
    """
    The action to execute with its specific parameters
    """

    type: typing.Literal["ai_upload_file"] = "ai_upload_file"
    selector: typing.Optional[str] = None
    file_url: typing.Optional[str] = None
    intention: typing.Optional[str] = None
    data: typing.Optional[UploadFileActionData] = None
    timeout: typing.Optional[float] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class RunSdkActionRequestAction_Extract(UniversalBaseModel):
    """
    The action to execute with its specific parameters
    """

    type: typing.Literal["extract"] = "extract"
    prompt: typing.Optional[str] = None
    extract_schema: typing.Optional[ExtractActionExtractSchema] = None
    error_code_mapping: typing.Optional[typing.Dict[str, typing.Optional[str]]] = None
    intention: typing.Optional[str] = None
    data: typing.Optional[ExtractActionData] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class RunSdkActionRequestAction_LocateElement(UniversalBaseModel):
    """
    The action to execute with its specific parameters
    """

    type: typing.Literal["locate_element"] = "locate_element"
    prompt: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class RunSdkActionRequestAction_Prompt(UniversalBaseModel):
    """
    The action to execute with its specific parameters
    """

    type: typing.Literal["prompt"] = "prompt"
    prompt: str
    response_schema: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = None
    model: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class RunSdkActionRequestAction_Validate(UniversalBaseModel):
    """
    The action to execute with its specific parameters
    """

    type: typing.Literal["validate"] = "validate"
    prompt: str
    model: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


RunSdkActionRequestAction = typing.Union[
    RunSdkActionRequestAction_AiAct,
    RunSdkActionRequestAction_AiClick,
    RunSdkActionRequestAction_AiInputText,
    RunSdkActionRequestAction_AiSelectOption,
    RunSdkActionRequestAction_AiUploadFile,
    RunSdkActionRequestAction_Extract,
    RunSdkActionRequestAction_LocateElement,
    RunSdkActionRequestAction_Prompt,
    RunSdkActionRequestAction_Validate,
]



from dotenv import load_dotenv

from skyvern.utils.env_paths import resolve_backend_env_path

from . import cli_app

if __name__ == "__main__":  # pragma: no cover - manual CLI invocation
    load_dotenv(resolve_backend_env_path())
    cli_app()



"""add queued_at started_at finished_at columns

Revision ID: add_run_timestamps
Revises: babaa7307e8a
Create Date: 2025-06-02 00:50:00+00:00
"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

revision: str = "add_run_timestamps"
down_revision: Union[str, None] = "babaa7307e8a"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.add_column("tasks", sa.Column("queued_at", sa.DateTime(), nullable=True))
    op.add_column("tasks", sa.Column("started_at", sa.DateTime(), nullable=True))
    op.add_column("tasks", sa.Column("finished_at", sa.DateTime(), nullable=True))

    op.add_column("observer_cruises", sa.Column("queued_at", sa.DateTime(), nullable=True))
    op.add_column("observer_cruises", sa.Column("started_at", sa.DateTime(), nullable=True))
    op.add_column("observer_cruises", sa.Column("finished_at", sa.DateTime(), nullable=True))

    op.add_column("workflow_runs", sa.Column("queued_at", sa.DateTime(), nullable=True))
    op.add_column("workflow_runs", sa.Column("started_at", sa.DateTime(), nullable=True))
    op.add_column("workflow_runs", sa.Column("finished_at", sa.DateTime(), nullable=True))


def downgrade() -> None:
    op.drop_column("workflow_runs", "finished_at")
    op.drop_column("workflow_runs", "started_at")
    op.drop_column("workflow_runs", "queued_at")

    op.drop_column("observer_cruises", "finished_at")
    op.drop_column("observer_cruises", "started_at")
    op.drop_column("observer_cruises", "queued_at")

    op.drop_column("tasks", "finished_at")
    op.drop_column("tasks", "started_at")
    op.drop_column("tasks", "queued_at")



# This file was auto-generated by Fern from our API Definition.

import typing

from .geo_target import GeoTarget
from .proxy_location import ProxyLocation

WorkflowRunProxyLocation = typing.Union[ProxyLocation, GeoTarget, typing.Dict[str, typing.Optional[typing.Any]]]



from skyvern.forge import app
from skyvern.forge.sdk.schemas.tasks import Task
from skyvern.schemas.runs import CUA_ENGINES, CUA_RUN_TYPES


async def is_cua_task(
    *,
    task: Task,
) -> bool:
    """Return True if the run, engine, or task indicates a CUA task."""

    if task.workflow_run_id:
        # it's a task based block, should look up the block run to see if it's a CUA task
        block = await app.DATABASE.get_workflow_run_block_by_task_id(
            task_id=task.task_id,
            organization_id=task.organization_id,
        )
        if block.engine is not None and block.engine in CUA_ENGINES:
            return True

    run = await app.DATABASE.get_run(
        run_id=task.task_id,
        organization_id=task.organization_id,
    )
    if run and run.task_run_type in CUA_RUN_TYPES:
        return True

    return False



import asyncio

from fastapi import Depends, HTTPException, Path, Query
from fastapi.responses import ORJSONResponse

from skyvern import analytics
from skyvern.forge import app
from skyvern.forge.sdk.routes.code_samples import (
    CLOSE_BROWSER_SESSION_CODE_SAMPLE_PYTHON,
    CLOSE_BROWSER_SESSION_CODE_SAMPLE_TS,
    CREATE_BROWSER_SESSION_CODE_SAMPLE_PYTHON,
    CREATE_BROWSER_SESSION_CODE_SAMPLE_TS,
    GET_BROWSER_SESSION_CODE_SAMPLE_PYTHON,
    GET_BROWSER_SESSION_CODE_SAMPLE_TS,
    GET_BROWSER_SESSIONS_CODE_SAMPLE_PYTHON,
    GET_BROWSER_SESSIONS_CODE_SAMPLE_TS,
)
from skyvern.forge.sdk.routes.routers import base_router
from skyvern.forge.sdk.schemas.organizations import Organization
from skyvern.forge.sdk.services import org_auth_service
from skyvern.schemas.browser_sessions import (
    CreateBrowserSessionRequest,
    ProcessBrowserSessionRecordingRequest,
    ProcessBrowserSessionRecordingResponse,
)
from skyvern.webeye.schemas import BrowserSessionResponse


@base_router.get(
    "/browser_sessions/history",
    include_in_schema=False,
)
@base_router.get(
    "/browser_sessions/history/",
    include_in_schema=False,
)
async def get_browser_sessions_all(
    current_org: Organization = Depends(org_auth_service.get_current_org),
    page: int = Query(1, ge=1, description="Page number for pagination"),
    page_size: int = Query(10, ge=1, le=100, description="Number of items per page"),
) -> list[BrowserSessionResponse]:
    """Get all browser sessions for the organization"""
    analytics.capture("skyvern-oss-agent-browser-sessions-get-all")

    browser_sessions = await app.DATABASE.get_persistent_browser_sessions_history(
        current_org.organization_id,
        page=page,
        page_size=page_size,
    )

    responses = await asyncio.gather(
        *[
            BrowserSessionResponse.from_browser_session(browser_session, app.STORAGE)
            for browser_session in browser_sessions
        ]
    )

    return responses


@base_router.post(
    "/browser_sessions",
    response_model=BrowserSessionResponse,
    tags=["Browser Sessions"],
    openapi_extra={
        "x-fern-sdk-method-name": "create_browser_session",
        "x-fern-examples": [
            {
                "code-samples": [
                    {"sdk": "python", "code": CREATE_BROWSER_SESSION_CODE_SAMPLE_PYTHON},
                    {"sdk": "typescript", "code": CREATE_BROWSER_SESSION_CODE_SAMPLE_TS},
                ]
            }
        ],
    },
    description="Create a browser session that persists across multiple runs",
    summary="Create a session",
    responses={
        200: {"description": "Successfully created browser session"},
        403: {"description": "Unauthorized - Invalid or missing authentication"},
    },
)
@base_router.post(
    "/browser_sessions/",
    response_model=BrowserSessionResponse,
    include_in_schema=False,
)
async def create_browser_session(
    browser_session_request: CreateBrowserSessionRequest = CreateBrowserSessionRequest(),
    current_org: Organization = Depends(org_auth_service.get_current_org),
) -> BrowserSessionResponse:
    browser_session = await app.PERSISTENT_SESSIONS_MANAGER.create_session(
        organization_id=current_org.organization_id,
        timeout_minutes=browser_session_request.timeout,
        proxy_location=browser_session_request.proxy_location,
        extensions=browser_session_request.extensions,
        browser_type=browser_session_request.browser_type,
    )
    return await BrowserSessionResponse.from_browser_session(browser_session)


@base_router.post(
    "/browser_sessions/{browser_session_id}/close",
    tags=["Browser Sessions"],
    openapi_extra={
        "x-fern-sdk-method-name": "close_browser_session",
        "x-fern-examples": [
            {
                "code-samples": [
                    {"sdk": "python", "code": CLOSE_BROWSER_SESSION_CODE_SAMPLE_PYTHON},
                    {"sdk": "typescript", "code": CLOSE_BROWSER_SESSION_CODE_SAMPLE_TS},
                ]
            }
        ],
    },
    description="Close a session. Once closed, the session cannot be used again.",
    summary="Close a session",
    responses={
        200: {"description": "Successfully closed browser session"},
        403: {"description": "Unauthorized - Invalid or missing authentication"},
    },
)
@base_router.post(
    "/browser_sessions/{browser_session_id}/close/",
    include_in_schema=False,
)
async def close_browser_session(
    browser_session_id: str = Path(
        ...,
        description="The ID of the browser session to close. completed_at will be set when the browser session is closed. browser_session_id starts with `pbs_`",
        examples=["pbs_123456"],
    ),
    current_org: Organization = Depends(org_auth_service.get_current_org),
) -> ORJSONResponse:
    await app.PERSISTENT_SESSIONS_MANAGER.close_session(current_org.organization_id, browser_session_id)
    return ORJSONResponse(
        content={"message": "Browser session closed"},
        status_code=200,
        media_type="application/json",
    )


@base_router.get(
    "/browser_sessions/{browser_session_id}",
    response_model=BrowserSessionResponse,
    tags=["Browser Sessions"],
    openapi_extra={
        "x-fern-sdk-method-name": "get_browser_session",
        "x-fern-examples": [
            {
                "code-samples": [
                    {"sdk": "python", "code": GET_BROWSER_SESSION_CODE_SAMPLE_PYTHON},
                    {"sdk": "typescript", "code": GET_BROWSER_SESSION_CODE_SAMPLE_TS},
                ]
            }
        ],
    },
    description="Get details about a specific browser session, including the browser address for cdp connection.",
    summary="Get a session",
    responses={
        200: {"description": "Successfully retrieved browser session details"},
        404: {"description": "Browser session not found"},
        403: {"description": "Unauthorized - Invalid or missing authentication"},
    },
)
@base_router.get(
    "/browser_sessions/{browser_session_id}/",
    response_model=BrowserSessionResponse,
    include_in_schema=False,
)
async def get_browser_session(
    browser_session_id: str = Path(
        ..., description="The ID of the browser session. browser_session_id starts with `pbs_`", examples=["pbs_123456"]
    ),
    current_org: Organization = Depends(org_auth_service.get_current_org),
) -> BrowserSessionResponse:
    analytics.capture("skyvern-oss-agent-browser-session-get")
    browser_session = await app.PERSISTENT_SESSIONS_MANAGER.get_session(
        browser_session_id,
        current_org.organization_id,
    )
    if not browser_session:
        raise HTTPException(status_code=404, detail=f"Browser session {browser_session_id} not found")
    return await BrowserSessionResponse.from_browser_session(browser_session, app.STORAGE)


@base_router.get(
    "/browser_sessions",
    response_model=list[BrowserSessionResponse],
    tags=["Browser Sessions"],
    openapi_extra={
        "x-fern-sdk-method-name": "get_browser_sessions",
        "x-fern-examples": [
            {
                "code-samples": [
                    {"sdk": "python", "code": GET_BROWSER_SESSIONS_CODE_SAMPLE_PYTHON},
                    {"sdk": "typescript", "code": GET_BROWSER_SESSIONS_CODE_SAMPLE_TS},
                ]
            }
        ],
    },
    description="Get all active browser sessions for the organization",
    summary="Get active browser sessions",
    responses={
        200: {"description": "Successfully retrieved all active browser sessions"},
        403: {"description": "Unauthorized - Invalid or missing authentication"},
    },
)
@base_router.get(
    "/browser_sessions/",
    response_model=list[BrowserSessionResponse],
    include_in_schema=False,
)
async def get_browser_sessions(
    current_org: Organization = Depends(org_auth_service.get_current_org),
) -> list[BrowserSessionResponse]:
    """Get all active browser sessions for the organization"""
    analytics.capture("skyvern-oss-agent-browser-sessions-get")
    browser_sessions = await app.PERSISTENT_SESSIONS_MANAGER.get_active_sessions(current_org.organization_id)
    return await asyncio.gather(
        *[
            BrowserSessionResponse.from_browser_session(browser_session, app.STORAGE)
            for browser_session in browser_sessions
        ]
    )


@base_router.post(
    "/browser_sessions/{browser_session_id}/process_recording",
    include_in_schema=False,
)
async def process_recording(
    browser_session_id: str = Path(..., description="The ID of the browser session.", examples=["pbs_123456"]),
    recording_request: ProcessBrowserSessionRecordingRequest = ProcessBrowserSessionRecordingRequest(),
    current_org: Organization = Depends(org_auth_service.get_current_org),
) -> ProcessBrowserSessionRecordingResponse:
    browser_session = await app.PERSISTENT_SESSIONS_MANAGER.get_session(
        browser_session_id,
        current_org.organization_id,
    )
    if not browser_session:
        raise HTTPException(status_code=404, detail=f"Browser session {browser_session_id} not found")

    blocks, parameters = await app.BROWSER_SESSION_RECORDING_SERVICE.process_recording(
        organization_id=current_org.organization_id,
        browser_session_id=browser_session_id,
        compressed_chunks=recording_request.compressed_chunks,
        workflow_permanent_id=recording_request.workflow_permanent_id,
    )

    return ProcessBrowserSessionRecordingResponse(blocks=blocks, parameters=parameters)



from __future__ import annotations

import json
import time
import typing

import structlog
from starlette.concurrency import iterate_in_threadpool

from skyvern.config import settings

if typing.TYPE_CHECKING:  # pragma: no cover - import only for type hints
    from typing import Awaitable, Callable

    from fastapi import Response
    from starlette.requests import Request

LOG = structlog.get_logger()

_SENSITIVE_HEADERS = {"authorization", "cookie", "x-api-key"}
_SENSITIVE_ENDPOINTS = {
    "POST /api/v1/credentials",
    "POST /v1/credentials",
    "POST /v1/credentials/onepassword/create",
    "POST /v1/credentials/azure_credential/create",
}
_MAX_BODY_LENGTH = 1000
_MAX_RESPONSE_READ_BYTES = 1024 * 1024  # 1 MB — skip logging bodies larger than this
_BINARY_PLACEHOLDER = "<binary>"
_REDACTED = "****"
_LOGGABLE_CONTENT_TYPES = {"text/", "application/json"}
_STREAMING_CONTENT_TYPE = "text/event-stream"

# Exact field names that are always redacted.  Use a set for O(1) lookup
# instead of regex substring matching to avoid false positives like
# credential_id, author, page_token, etc.
_SENSITIVE_FIELDS: set[str] = {
    "password",
    "secret",
    "token",
    "api_key",
    "apikey",
    "api-key",
    "credential",
    "access_key",
    "private_key",
    "auth",
    "authorization",
    "secret_key",
}


def _sanitize_headers(headers: typing.Mapping[str, str]) -> dict[str, str]:
    sanitized: dict[str, str] = {}
    for key, value in headers.items():
        if key.lower() in _SENSITIVE_HEADERS:
            continue
        sanitized[key] = value
    return sanitized


def _sanitize_body(request: Request, body: bytes, content_type: str | None) -> str:
    if f"{request.method.upper()} {request.url.path.rstrip('/')}" in _SENSITIVE_ENDPOINTS:
        return _REDACTED
    if not body:
        return ""
    if content_type and not (content_type.startswith("text/") or content_type.startswith("application/json")):
        return _BINARY_PLACEHOLDER
    try:
        text = body.decode("utf-8", errors="replace")
    except Exception:
        return _BINARY_PLACEHOLDER
    if len(text) > _MAX_BODY_LENGTH:
        return text[:_MAX_BODY_LENGTH] + "...[truncated]"
    return text


def _is_sensitive_key(key: str) -> bool:
    return key.lower() in _SENSITIVE_FIELDS


def _redact_sensitive_fields(obj: typing.Any, _depth: int = 0) -> typing.Any:
    """Redact dict values whose *key name* exactly matches a known sensitive field.

    Uses exact-match (case-insensitive) rather than substring/regex to avoid
    false positives on fields like ``credential_id``, ``author``, or
    ``page_token`` which contain sensitive substrings but are not secrets.
    """
    if _depth > 20:
        # Stop recursing but still redact sensitive keys at this level
        if isinstance(obj, dict):
            return {k: _REDACTED if _is_sensitive_key(k) else v for k, v in obj.items()}
        return obj
    if isinstance(obj, dict):
        return {
            k: _REDACTED if _is_sensitive_key(k) else _redact_sensitive_fields(v, _depth + 1) for k, v in obj.items()
        }
    if isinstance(obj, list):
        return [_redact_sensitive_fields(item, _depth + 1) for item in obj]
    return obj


def _is_loggable_content_type(content_type: str | None) -> bool:
    if not content_type:
        return True  # assume text when header is missing
    return any(content_type.startswith(prefix) for prefix in _LOGGABLE_CONTENT_TYPES)


def _sanitize_response_body(request: Request, body_str: str | None, content_type: str | None) -> str:
    if f"{request.method.upper()} {request.url.path.rstrip('/')}" in _SENSITIVE_ENDPOINTS:
        return _REDACTED
    if body_str is None:
        return _BINARY_PLACEHOLDER
    if not body_str:
        return ""
    if not _is_loggable_content_type(content_type):
        return _BINARY_PLACEHOLDER
    try:
        parsed = json.loads(body_str)
        redacted = _redact_sensitive_fields(parsed)
        text = json.dumps(redacted)
    except (json.JSONDecodeError, TypeError):
        text = body_str
    if len(text) > _MAX_BODY_LENGTH:
        return text[:_MAX_BODY_LENGTH] + "...[truncated]"
    return text


async def _get_response_body_str(response: Response) -> str | None:
    """Read and reconstitute the response body for logging.

    Returns ``None`` when the body is binary or exceeds
    ``_MAX_RESPONSE_READ_BYTES`` to avoid buffering large payloads
    solely for logging purposes.
    """
    response_body = b""
    async for chunk in response.body_iterator:
        response_body += chunk
    response.body_iterator = iterate_in_threadpool(iter([response_body]))

    if len(response_body) > _MAX_RESPONSE_READ_BYTES:
        return None

    try:
        return response_body.decode("utf-8")
    except UnicodeDecodeError:
        return None


async def log_raw_request_middleware(request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
    if not settings.LOG_RAW_API_REQUESTS:
        return await call_next(request)

    start_time = time.monotonic()
    body_bytes = await request.body()
    # ensure downstream handlers can access body again
    try:
        request._body = body_bytes  # type: ignore[attr-defined]
    except Exception:
        pass

    url_path = request.url.path
    http_method = request.method
    sanitized_headers = _sanitize_headers(dict(request.headers))
    body_text = _sanitize_body(request, body_bytes, request.headers.get("content-type"))

    try:
        response = await call_next(request)

        if response.status_code >= 500:
            log_method = LOG.error
        elif response.status_code >= 400:
            log_method = LOG.warning
        else:
            log_method = LOG.info

        resp_content_type = response.headers.get("content-type", "")
        if _STREAMING_CONTENT_TYPE in resp_content_type:
            response_body = "<streaming>"
        else:
            raw_response_body = await _get_response_body_str(response)
            response_body = _sanitize_response_body(request, raw_response_body, resp_content_type)

        log_method(
            "api.raw_request",
            method=http_method,
            path=url_path,
            status_code=response.status_code,
            body=body_text,
            headers=sanitized_headers,
            response_body=response_body,
            # backwards-compat: keep error_body for existing Datadog queries
            error_body=response_body if response.status_code >= 400 else None,
            duration_seconds=time.monotonic() - start_time,
        )
        return response
    except Exception:
        LOG.error(
            "api.raw_request",
            method=http_method,
            path=url_path,
            body=body_text,
            headers=sanitized_headers,
            exc_info=True,
            duration_seconds=time.monotonic() - start_time,
        )
        raise



import typing as t

import structlog

from skyvern.services.browser_recording.types import (
    Action,
    ActionInputText,
    ActionKind,
    ActionTarget,
    EventTarget,
    ExfiltratedEvent,
    Mouse,
    MousePosition,
    target_has_changed,
)

from .state_machine import StateMachine

LOG = structlog.get_logger()


class StateMachineInputText(StateMachine):
    state: t.Literal["focus", "keydown", "blur"] = "focus"
    target: EventTarget | None = None
    timestamp_start: float | None = None
    mouse: MousePosition | None = None

    def __init__(self) -> None:
        self.reset()

    def tick(self, event: ExfiltratedEvent, current_actions: list[Action]) -> ActionInputText | None:
        if event.source != "console":
            return None

        match self.state:
            case "focus":
                if event.params.type != "focus":
                    if event.params.mousePosition:
                        if event.params.mousePosition.xp is not None and event.params.mousePosition.yp is not None:
                            self.mouse = event.params.mousePosition

                    return None

                LOG.debug(f"~ focus detected [{event.params.target.skyId or event.params.target.id}]")

                if event.params.type == "focus":
                    self.target = event.params.target
                    self.state = "keydown"
                    self.timestamp_start = event.params.timestamp

            case "keydown":
                if event.params.type != "keydown":
                    return None

                if self.update_target(event.params.target):
                    return None

                LOG.debug(f"~ initial keydown detected '{event.params.key}'")

                self.state = "blur"
            case "blur":
                if event.params.type == "keydown":
                    LOG.debug(f"~ input-text: subsequent keydown detected '{event.params.key}'")

                    if event.params.key == "Enter":
                        return self.emit(event)

                    return None

                if event.params.type != "blur":
                    return None

                if self.update_target(event.params.target):
                    return None

                LOG.debug("~ blur detected")

                return self.emit(event)

        return None

    def update_target(self, event_target: EventTarget) -> bool:
        if target_has_changed(self.target, event_target):
            self.target = event_target

            LOG.debug("~ input-text target changed, resetting state machine")

            self.reset()

            return True

        return False

    def emit(self, event: ExfiltratedEvent) -> ActionInputText | None:
        if not self.target:
            LOG.debug("~ cannot emit, missing target or mouse; resetting")

            self.reset()
            return None

        xp = (self.mouse.xp or -1) if self.mouse else None
        yp = (self.mouse.yp or -1) if self.mouse else None

        LOG.debug("~ emitting input text action", exfiltrated_event=event)

        input_value = event.params.target.value

        if input_value is None:
            LOG.debug("~ cannot emit, missing input value; resetting")

            self.reset()
            return None

        action_target = ActionTarget(
            class_name=self.target.className,
            id=self.target.id,
            mouse=Mouse(xp=xp, yp=yp),
            sky_id=self.target.skyId,
            tag_name=self.target.tagName,
            texts=self.target.text,
        )

        action = ActionInputText(
            kind=ActionKind.INPUT_TEXT.value,
            target=action_target,
            timestamp_start=self.timestamp_start,
            timestamp_end=event.params.timestamp,
            url=event.params.url,
            input_value=str(input_value),
        )

        self.reset()

        return action

    def on_action(self, action: Action, current_actions: list[Action]) -> bool:
        if action.kind == ActionKind.CLICK:
            # NOTE(jdo): skipping self.reset here; a focus event on an element can often be followed by a
            # click event, and the identity doesn't always match due to nesting. I think a more precise
            # check would be to:
            #  - check identity match for click and focus; if not matching:
            #   - ask the browser via cdp if there is a nesting relation between the two elements
            #   - if yes, allow and carry on, otherwise reset
            # That's another round trip. It's likely pretty fast, tho. For now, we'll just assume a click does
            # not invalidate the state of the input text state machine.
            return True

        self.reset()

        return True

    def reset(self) -> None:
        self.state = "focus"
        self.target = None
        self.mouse = None
        self.timestamp_start = None



"""
A channel for executing JavaScript against a persistent browser instance.

What this channel looks like:

    [API Server] <--> [Browser (CDP)]

Channel data:

    Chrome DevTools Protocol (CDP) over WebSockets. We cheat and use Playwright.
"""

from __future__ import annotations

import typing as t
from contextlib import asynccontextmanager

import structlog
from playwright.async_api import Browser, BrowserContext, Page, Playwright

from skyvern.forge.sdk.routes.streaming.channels.cdp import CdpChannel

if t.TYPE_CHECKING:
    from skyvern.forge.sdk.routes.streaming.channels.vnc import VncChannel

LOG = structlog.get_logger()


class ExecutionChannel(CdpChannel):
    # Explicitly declare inherited attributes for mypy when follow_imports = skip
    browser: Browser | None
    browser_context: BrowserContext | None
    page: Page | None
    pw: Playwright | None
    """
    ExecutionChannel.
    """

    @property
    def class_name(self) -> str:
        return self.__class__.__name__

    async def get_selected_text(self) -> str:
        LOG.info(f"{self.class_name} getting selected text", **self.identity)

        js_expression = """
        () => {
            const selection = window.getSelection();
            return selection ? selection.toString() : '';
        }
        """

        selected_text = await self.evaluate_js(js_expression)

        if isinstance(selected_text, str) or selected_text is None:
            LOG.info(
                f"{self.class_name} got selected text",
                length=len(selected_text) if selected_text else 0,
                **self.identity,
            )
            return selected_text or ""

        raise RuntimeError(f"{self.class_name} selected text is not a string, but a(n) '{type(selected_text)}'")

    async def paste_text(self, text: str) -> None:
        LOG.info(f"{self.class_name} pasting text", **self.identity)

        js_expression = """
        (text) => {
            const activeElement = document.activeElement;
            if (activeElement && (activeElement.tagName === 'INPUT' || activeElement.tagName === 'TEXTAREA' || activeElement.isContentEditable)) {
                const start = activeElement.selectionStart || 0;
                const end = activeElement.selectionEnd || 0;
                const value = activeElement.value || '';
                activeElement.value = value.slice(0, start) + text + value.slice(end);
                const newCursorPos = start + text.length;
                activeElement.setSelectionRange(newCursorPos, newCursorPos);
            }
        }
        """

        await self.evaluate_js(js_expression, text)

        LOG.info(f"{self.class_name} pasted text successfully", **self.identity)

    async def close(self) -> None:
        LOG.info(f"{self.class_name} closing connection", **self.identity)

        if self.browser:
            await self.browser.close()
            self.browser = None
            self.browser_context = None
            self.page = None

        if self.pw:
            await self.pw.stop()
            self.pw = None

        LOG.info(f"{self.class_name} closed", **self.identity)


@asynccontextmanager
async def execution_channel(vnc_channel: VncChannel) -> t.AsyncIterator[ExecutionChannel]:
    """
    The first pass at this has us doing the following for every operation:
      - creating a new channel
      - connecting
      - [doing smth]
      - closing the channel

    This may add latency, but locally it is pretty fast. This keeps things stateless for now.

    If it turns out it's too slow, we can refactor to keep a persistent channel per vnc client.
    """

    channel = ExecutionChannel(vnc_channel=vnc_channel)

    try:
        await channel.connect()

        # NOTE(jdo:streaming-local-dev)
        # from skyvern.config import settings
        # await channel.connect(settings.BROWSER_REMOTE_DEBUGGING_URL)

        yield channel
    finally:
        await channel.close()



"""Add model column to workflows table

Revision ID: af49ca791fc7
Revises: ecdba16ef45c
Create Date: 2025-05-29 14:35:38.232668+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "af49ca791fc7"
down_revision: Union[str, None] = "ecdba16ef45c"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflows", sa.Column("model", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflows", "model")
    # ### end Alembic commands ###



"""add cache_key to workflows table

Revision ID: f2e78df26c97
Revises: dd29417b397c
Create Date: 2025-08-06 07:04:57.428538+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "f2e78df26c97"
down_revision: Union[str, None] = "dd29417b397c"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflows", sa.Column("cache_key", sa.String(), nullable=True))
    op.drop_column("workflows", "cache_project_id")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflows", sa.Column("cache_project_id", sa.VARCHAR(), autoincrement=False, nullable=True))
    op.drop_column("workflows", "cache_key")
    # ### end Alembic commands ###



"""add script_blocks.input_fields

Revision ID: 7ab8e817802a
Revises: cf6ae2f5013c
Create Date: 2025-12-10 23:56:33.299511+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "7ab8e817802a"
down_revision: Union[str, None] = "cf6ae2f5013c"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("script_blocks", sa.Column("input_fields", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("script_blocks", "input_fields")
    # ### end Alembic commands ###



# This file was auto-generated by Fern from our API Definition.

import typing

BlockType = typing.Union[
    typing.Literal[
        "task",
        "task_v2",
        "for_loop",
        "conditional",
        "code",
        "text_prompt",
        "download_to_s3",
        "upload_to_s3",
        "file_upload",
        "send_email",
        "file_url_parser",
        "validation",
        "action",
        "navigation",
        "extraction",
        "login",
        "wait",
        "file_download",
        "goto_url",
        "pdf_parser",
        "http_request",
        "human_interaction",
        "print_page",
    ],
    typing.Any,
]
